"""User-interface related functions for building bots.

This module supports several different bot classes which could be used in
conjunction. Each bot should subclass at least one of these four classes:

* :class:`BaseBot`: Basic bot class in case where the site is handled
  differently, like working on multiple sites in parallel. No site
  attribute is provided. Instead site of the current page should be used.
  This class should normally not be used directly.

* :class:`SingleSiteBot`: Bot class which should only be run on a single
  site. They usually store site specific content and thus can't be
  easily run when the generator returns a page on another site. It has a
  property ``site`` which can also be changed. If the generator returns
  a page of a different site it'll skip that page.

* :class:`MultipleSitesBot`: An alias of :class:`BaseBot`. Should not
  be used if any other bot class is used.

* :class:`ConfigParserBot`: Bot class which supports reading options
  from a ``scripts.ini`` configuration file. That file consists of
  sections, led by a ``[section]`` header and followed by
  ``option: value`` or ``option=value`` entries. The section is the
  script name without .py suffix. All options identified must be
  predefined in available_options dictionary.

* :class:`Bot`: The previous base class which should be avoided. This
  class is mainly used for bots which work with Wikibase or together
  with an image repository.

Additionally there is the :py:obj:`CurrentPageBot` class which
automatically sets the current page to the page treated. It is
recommended to use this class and to use ``treat_page`` instead of
``treat`` and ``put_current`` instead of ``userPut``. It by default
subclasses the ``BaseBot`` class.

With :class:`CurrentPageBot` it's possible to subclass one of the
following classes to filter the pages which are ultimately handled by
:meth:`CurrentPageBot.treat_page`:

* :class:`ExistingPageBot`: Only handle pages which do exist.
* :class:`CreatingPageBot`: Only handle pages which do not exist.
* :class:`FollowRedirectPageBot`: If the generator returns a redirect
  page it'll follow the redirect and instead work on the redirected
  class.

It is possible to combine filters by subclassing multiple of them. They
are new-style classes so when a class is first subclassing
:class:`ExistingPageBot` and then :class:`FollowRedirectPageBot` it
will also work on pages which do not exist when a redirect pointed to
that. If the order is inversed it'll first follow them and then check
whether they exist.

Additionally there is the :class:`AutomaticTWSummaryBot` which
subclasses :class:`CurrentPageBot` and automatically defines the summary
when :meth:`put_current` is used.

.. deprecated:: 9.2
   The functions
   :func:`critical()<pywikibot.logging.critical>`
   :func:`debug()<pywikibot.logging.debug>`
   :func:`error()<pywikibot.logging.error>`
   :func:`exception()<pywikibot.logging.exception>`
   :func:`log()<pywikibot.logging.log>`
   :func:`output()<pywikibot.logging.output>`
   :func:`stdout()<pywikibot.logging.stdout>` and
   :func:`warning()<pywikibot.logging.warning>`
   as well as the constants
   ``CRITICAL``, ``DEBUG``, ``ERROR``, ``INFO``,
   :attr:`INPUT<pywikibot.logging.INPUT>`
   :attr:`STDOUT<pywikibot.logging.STDOUT>`
   :attr:`VERBOSE<pywikibot.logging.VERBOSE>`
   and ``WARNING`` imported from :mod:`logging` module are deprecated
   within this module. Import them directly. These functions can also be
   used as :mod:`pywikibot` members.

.. versionremoved:: 10.0
   The bot classes :class:`RedirectPageBot` and
   :class:`NoRedirectPageBot` are deprecated. Use
   :attr:`use_redirects<BaseBot.use_redirects>` attribute instead.
"""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations


__all__ = (
    'LoggingFormatter',
    'set_interface', 'init_handlers', 'writelogheader',
    'input', 'input_choice', 'input_yn', 'input_list_choice', 'ui',
    'Option', 'StandardOption', 'NestedOption', 'IntegerOption',
    'ContextOption', 'ListOption', 'ShowingListOption', 'MultipleChoiceList',
    'ShowingMultipleChoiceList', 'OutputProxyOption',
    'HighlightContextOption', 'ChoiceException', 'UnhandledAnswer',
    'Choice', 'StaticChoice', 'LinkChoice', 'AlwaysChoice',
    'QuitKeyboardInterrupt',
    'InteractiveReplace',
    'calledModuleName', 'handle_args',
    'show_help', 'suggest_help',
    'writeToCommandLogFile', 'open_webbrowser',
    'OptionHandler',
    'BaseBot', 'Bot', 'ConfigParserBot', 'SingleSiteBot', 'MultipleSitesBot',
    'CurrentPageBot', 'AutomaticTWSummaryBot',
    'ExistingPageBot', 'FollowRedirectPageBot', 'CreatingPageBot',
    'WikidataBot',
)

import atexit
import configparser
import json
import logging
import logging.handlers
import os
import re
import sys
import time
import warnings
import webbrowser
from collections import Counter
from collections.abc import Container, Generator
from functools import wraps
from importlib import import_module
from pathlib import Path
from textwrap import fill
from typing import TYPE_CHECKING, Any

import pywikibot
import pywikibot.logging as pwb_logging
from pywikibot import config, daemonize, i18n, version
from pywikibot.backports import Callable, Dict, Iterable, Sequence
from pywikibot.bot_choice import (
    AlwaysChoice,
    Choice,
    ChoiceException,
    ContextOption,
    HighlightContextOption,
    IntegerOption,
    InteractiveReplace,
    LinkChoice,
    ListOption,
    MultipleChoiceList,
    NestedOption,
    Option,
    OutputProxyOption,
    QuitKeyboardInterrupt,
    ShowingListOption,
    ShowingMultipleChoiceList,
    StandardOption,
    StaticChoice,
    UnhandledAnswer,
)
from pywikibot.exceptions import (
    EditConflictError,
    Error,
    LockedPageError,
    NoPageError,
    PageSaveRelatedError,
    ServerError,
    SpamblacklistError,
    UnknownFamilyError,
    UnknownSiteError,
    VersionParseError,
    WikiBaseError,
)

# the constants are
from pywikibot.logging import (  # noqa: F401
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    INPUT,
    STDOUT,
    VERBOSE,
    WARNING,
    add_init_routine,
)
from pywikibot.logging import debug as _debug
from pywikibot.logging import error as _error
from pywikibot.logging import exception as _exception
from pywikibot.logging import log as _log
from pywikibot.logging import stdout as _stdout
from pywikibot.logging import warning as _warning
from pywikibot.throttle import Throttle
from pywikibot.tools import redirect_func, strtobool
from pywikibot.tools._logging import LoggingFormatter


if TYPE_CHECKING:
    from typing import Union

    from pywikibot.site import BaseSite

    AnswerType = Union[Iterable[Union[tuple[str, str], Option]], Option]

_GLOBAL_HELP = """
GLOBAL OPTIONS
==============
(Global arguments available for all bots)

-dir:PATH         Read the bot's configuration data from directory given
                  by PATH, instead of from the default directory.

-config:xyn       The user config filename. Default is user-config.py.

-lang:xx          Set the language of the wiki you want to work on,
                  overriding the configuration in user config file.
                  xx should be the site code.

-family:xyz       Set the family of the wiki you want to work on, e.g.
                  wikipedia, wiktionary, wikivoyage, ... This will
                  override the configuration in user config file.

-site:xyz:xx      Set the wiki site you want to work on, e.g.
                  wikipedia:test, wiktionary:de, wikivoyage:en, ... This
                  will override the configuration in user config file.

-user:xyz         Log in as user 'xyz' instead of the default username.

-daemonize:xyz    Immediately return control to the terminal and redirect
                  stdout and stderr to file xyz.
                  (only use for bots that require no input from stdin).

-help             Show this help text.

-log              Enable the log file, using the default filename
                  '{}-bot.log'
                  Logs will be stored in the logs subdirectory.

-log:xyz          Enable the log file, using 'xyz' as the filename.

-nolog            Disable the log file (if it is enabled by default).
                  Also disable command.log.

-maxlag           Sets a new maxlag parameter to a number of seconds.
                  Defer bot edits during periods of database server lag.
                  Default is set by config.py

-putthrottle:n    Set the minimum time (in seconds) the bot will wait
-pt:n             between saving pages.
-put_throttle:n

-debug:item       Enable the log file and include extensive debugging
-debug            data for component "item" (for all components if the
                  second form is used).

-verbose          Have the bot provide additional console output that may be
-v                useful in debugging.

-cosmeticchanges  Toggles the cosmetic_changes setting made in config.py
-cc               or user config file to its inverse and overrules it.
                  All other settings and restrictions are untouched. The
                  setting may also be given directly like `-cc:True`;
                  accepted values for the option are `1`, `yes`, `true`,
                  `on`, `y`, `t` for True and `0`, `no`, `false`, `off`,
                  `n`, `f` for False. Values are case-insensitive.

-simulate         Disables writing to the server. Useful for testing and
                  debugging of new code (if given, doesn't do any real
                  changes, but only shows what would have been changed).
                  An integer or float value may be given to simulate a
                  processing time; the bot just waits for given seconds.

-<config var>:n   You may use all given numeric config variables as
                  option and modify it with command line.

"""

_GLOBAL_HELP_NOTE = """
GLOBAL OPTIONS
==============
For global options use -help:global or run pwb.py -help

"""

ui: pywikibot.userinterfaces._interface_base.ABUIC | None = None
"""Holds a user interface object defined in :mod:`pywikibot.userinterfaces`
subpackage."""

#: global args used by tests via pwb wrapper
global_args: list[str] | None = None


def set_interface(module_name: str) -> None:
    """Configures any bots to use the given interface module.

    Search for user interface module in the
    :mod:`pywikibot.userinterfaces` subdirectory and initialize UI.
    Calls :func:`init_handlers` to re-initialize if we were already
    initialized with another UI.

    .. versionadded:: 6.4
    """
    global ui

    ui_module = __import__(f'pywikibot.userinterfaces.{module_name}_interface',
                           fromlist=['UI'])
    ui = ui_module.UI()
    assert ui is not None
    atexit.register(ui.flush)
    pywikibot.argvu = ui.argvu()

    # re-initialize
    if _handlers_initialized:
        _handlers_initialized.clear()
        init_handlers()


_handlers_initialized = []  # we can have a script and the script wrapper


def handler_namer(name: str) -> str:
    """Modify the filename of a log file when rotating.

    RotatingFileHandler will save old log files by appending the
    extensions ``.1``, ``.2`` etc., to the filename. To keep the
    original extension, which is usually ``.log``, this function
    swaps the appended counter with the log extension:

    >>> handler_namer('add_text.log.1')
    'add_text.1.log'

    .. versionadded:: 6.5
    """
    path, qualifier = name.rsplit('.', 1)
    root, ext = os.path.splitext(path)
    return f'{root}.{qualifier}{ext}'


def init_handlers() -> None:
    """Initialize the handlers and formatters for the logging system.

    This relies on the global variable :attr:`ui` which is a UI object.

    .. seealso:: :mod:`pywikibot.userinterfaces`

    Calls :func:`writelogheader` after handlers are initialized.
    This function must be called before using any input/output methods;
    and must be called again if ui handler is changed. Use
    :func:`set_interface` to set the new interface which initializes it.

    .. note:: this function is called by any user input and output
       function, so it should normally not need to be called explicitly.

    All user output is routed through the logging module.
    Each type of output is handled by an appropriate handler object.
    This structure is used to permit eventual development of other
    user interfaces (GUIs) without modifying the core bot code.

    The following output levels are defined:

     - DEBUG: only for file logging; debugging messages.
     - STDOUT: output that must be sent to sys.stdout (for bots that may
       have their output redirected to a file or other destination).
     - VERBOSE: optional progress information for display to user.
     - INFO: normal (non-optional) progress information for display to user.
     - INPUT: prompts requiring user response.
     - WARN: user warning messages.
     - ERROR: user error messages.
     - CRITICAL: fatal error messages.

     .. seealso::
        * :mod:`pywikibot.logging`
        * :python:`Python Logging Levels<library/logging.html#logging-levels>`

    Accordingly, do **not** use print statements in bot code; instead,
    use :func:`pywikibot.info()<pywikibot.logging.info>` function and
    other functions from :mod:`pywikibot.logging` module.

    .. versionchanged:: 6.2
       Different logfiles are used if multiple processes of the same
       script are running.
    """
    module_name = calledModuleName()
    if not module_name:
        module_name = 'terminal-interface'

    logging.addLevelName(VERBOSE, 'VERBOSE')
    # for messages to be displayed on terminal at "verbose" setting
    # use INFO for messages to be displayed even on non-verbose setting

    logging.addLevelName(STDOUT, 'STDOUT')
    # for messages to be displayed to stdout

    logging.addLevelName(INPUT, 'INPUT')
    # for prompts requiring user response

    root_logger = logging.getLogger('pywiki')
    if root_logger.hasHandlers() and module_name in _handlers_initialized:
        return
    root_logger.setLevel(DEBUG + 1)  # all records except DEBUG go to logger

    warnings_logger = logging.getLogger('py.warnings')
    warnings_logger.setLevel(DEBUG)

    # If there are command line warnings options, do not override them
    if not sys.warnoptions:
        logging.captureWarnings(True)

        if config.debug_log or 'deprecation' in config.log:
            warnings.filterwarnings('always')
        elif config.verbose_output:
            warnings.filterwarnings('module')
        warnings.filterwarnings('once', category=FutureWarning)

    for handler in root_logger.handlers:
        handler.close()
    root_logger.handlers.clear()  # remove any old handlers
    root_logger.propagate = False  # T281643

    # configure handler(s) for display to user interface
    assert ui is not None
    ui.init_handlers(root_logger, **config.userinterface_init_kwargs)

    # if user has enabled file logging, configure file handler
    if module_name in config.log or '*' in config.log:
        # Use a dummy Throttle to get a PID.
        # This is necessary because tests may have site disabled.
        throttle = Throttle('')
        pid_int = throttle.get_pid(module_name)  # get the global PID
        pid = str(pid_int) + '-' if pid_int > 1 else ''

        if config.logfilename:
            # keep config.logfilename unchanged
            logfile = config.datafilepath('logs', config.logfilename)
        else:
            # add PID to logfle name
            logfile = config.datafilepath('logs',
                                          f'{module_name}-{pid}bot.log')

        file_handler = logging.handlers.RotatingFileHandler(
            filename=logfile,
            maxBytes=config.logfilesize << 10,
            backupCount=config.logfilecount,
            encoding='utf-8'
        )
        file_handler.namer = handler_namer

        file_handler.setLevel(DEBUG)
        form = LoggingFormatter(
            fmt='%(asctime)s %(caller_file)18s, %(caller_line)4s '
                'in %(caller_name)18s: %(levelname)-8s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
        file_handler.setFormatter(form)
        root_logger.addHandler(file_handler)
        # Turn on debugging for each component requested by user
        # or for all components if nothing was specified
        for component in config.debug_log:
            if component:
                debuglogger = logging.getLogger('pywiki.' + component)
            else:
                debuglogger = logging.getLogger('pywiki')
            debuglogger.setLevel(DEBUG)
            debuglogger.addHandler(file_handler)

        warnings_logger.addHandler(file_handler)

    _handlers_initialized.append(module_name)

    writelogheader()


def writelogheader() -> None:
    """Save additional version, system and status info to the log file in use.

    This may help the user to track errors or report bugs.

    .. versionchanged:: 9.0
       ignore milliseconds with timestamp.
    """
    _log(f'\n=== Pywikibot framework v{pywikibot.__version__} --'
         ' Logging header ===')

    # script call
    _log(f'COMMAND: {sys.argv}')

    # script call time stamp
    _log(f'DATE: {pywikibot.Timestamp.nowutc()} UTC')

    # new framework release/revision? (handle_args needs to be called first)
    try:
        _log('VERSION: {}'.format(version.getversion(
            online=config.log_pywiki_repo_version).strip()))
    except VersionParseError:
        _exception()

    # system
    if hasattr(os, 'uname'):
        _log(f'SYSTEM: {os.uname()}')

    # config file dir
    _log(f'CONFIG FILE DIR: {pywikibot.config.base_dir}')

    # These are the main dependencies of pywikibot.
    check_package_list = [
        'requests',
        'mwparserfromhell',
    ]

    # report all imported packages
    if config.verbose_output:
        check_package_list += sys.modules

    _log('PACKAGES:')
    packages = version.package_versions(check_package_list)
    for name in sorted(packages.keys()):
        info = packages[name]
        info.setdefault('path',
                        f"[{info.get('type', 'path unknown')}]")
        info.setdefault('ver', '??')
        if 'err' in info:
            _log('  {name}: {err}'.format_map(info))
        else:
            _log('  {name} ({path}) = {ver}'.format_map(info))

    # imported modules
    _log('MODULES:')
    for module in sys.modules.copy().values():
        filename = version.get_module_filename(module)
        if not filename:
            continue

        mtime = version.get_module_mtime(module).isoformat(sep=' ',
                                                           timespec='seconds')
        _log(f'  {mtime} {filename}')

    if config.log_pywiki_repo_version:
        _log(f'PYWIKI REPO VERSION: {version.getversion_onlinerepo()}')

    _log('=' * 57)


add_init_routine(init_handlers)


# User input functions

def initialize_handlers(function):
    """Make sure logging system has been initialized.

    .. versionadded:: 7.0
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        init_handlers()
        return function(*args, **kwargs)
    return wrapper


@initialize_handlers
def input(question: str,
          password: bool = False,
          default: str | None = '',
          force: bool = False) -> str:
    """Ask the user a question, return the user's answer.

    :param question: a string that will be shown to the user. Don't add
        a space after the question mark/colon, this method will do this
        for you.
    :param password: if True, hides the user's input (for password
        entry).
    :param default: The default answer if none was entered. None to
        require an answer.
    :param force: Automatically use the default
    """
    assert ui is not None
    return ui.input(question, password=password, default=default, force=force)


@initialize_handlers
def input_choice(question: str,
                 answers: AnswerType,
                 default: str | None = None,
                 return_shortcut: bool = True,
                 automatic_quit: bool = True,
                 force: bool = False) -> Any:
    """Ask the user the question and return one of the valid answers.

    .. seealso::
       * :meth:`userinterfaces._interface_base.ABUIC.input_choice`
       * :meth:`userinterfaces.buffer_interface.UI.input_choice`
       * :meth:`userinterfaces.terminal_interface_base.UI.input_choice`

    :param question: The question asked without trailing spaces.
    :param answers: The valid answers each containing a full length answer and
        a shortcut. Each value must be unique.
    :param default: The result if no answer was entered. It must not be in the
        valid answers and can be disabled by setting it to None. If it should
        be linked with the valid answers it must be its shortcut.
    :param return_shortcut: Whether the shortcut or the index of the answer is
        returned.
    :param automatic_quit: Adds the option 'Quit' ('q') and throw a
        :exc:`bot.QuitKeyboardInterrupt` if selected.
    :param force: Automatically use the *default*.
    :return: The selected answer shortcut or index. Is -1 if the default is
        selected, it does not return the shortcut and the default is not a
        valid shortcut.
    """
    assert ui is not None
    return ui.input_choice(question, answers, default, return_shortcut,
                           automatic_quit=automatic_quit, force=force)


def input_yn(question: str,
             default: bool | str | None = None,
             automatic_quit: bool = True,
             force: bool = False) -> bool:
    """Ask the user a yes/no question and return the answer as a bool.

    **Example:**

    >>> input_yn('Do you like Pywikibot?', 'y', False, force=True)
    ... # doctest: +SKIP
    Do you like Pywikibot? ([Y]es, [n]o)
    True
    >>> input_yn('More examples?', False, automatic_quit=False, force=True)
    ... # doctest: +SKIP
    Some more examples? ([y]es, [N]o)
    False

    .. seealso:: :func:`input_choice`

    :param question: The question asked without trailing spaces.
    :param default: The result if no answer was entered. It must be a
        bool or ``'y'``, ``'n'``, ``0`` or ``1`` and can be disabled by
        setting it to None.
    :param automatic_quit: Adds the option 'Quit' ('q') and throw a
        :exc:`bot.QuitKeyboardInterrupt` if selected.
    :param force: Automatically use the *default*.
    :return: Return True if the user selected yes and False if the user
        selected no. If the default is not None it'll return True if default
        is True or 'y' and False if default is False or 'n'.
    """
    if isinstance(default, bool):
        default = 'ny'[default]

    return input_choice(question, [('Yes', 'y'), ('No', 'n')],
                        default,
                        automatic_quit=automatic_quit, force=force) == 'y'


@initialize_handlers
def input_list_choice(question: str,
                      answers: AnswerType,
                      default: int | str | None = None,
                      force: bool = False) -> str:
    """Ask the user the question and return one of the valid answers.

    :param question: The question asked without trailing spaces.
    :param answers: The valid answers each containing a full length
        answer.
    :param default: The result if no answer was entered. It must not be
        in the valid answers and can be disabled by setting it to None.
    :param force: Automatically use the default
    :return: The selected answer.
    """
    assert ui is not None
    return ui.input_list_choice(question, answers, default=default,
                                force=force)


# Command line parsing and help
def calledModuleName() -> str:
    """Return the name of the module calling this function.

    This is required because the -help option loads the module's
    docstring and because the module name will be used for the filename
    of the log.

    .. versionchanged:: 10.0
       Detect unittest and pytest run and return the test module.
    """
    mod = pywikibot.argvu[0]

    if 'pytest' in mod:
        return 'pytest'

    if mod.endswith('unittest'):
        return 'unittest'

    return Path(mod).stem


def handle_args(args: Iterable[str] | None = None,
                do_help: bool = True) -> list[str]:
    """Handle global command line arguments and return the rest as a list.

    Takes the command line arguments as strings, processes all
    :ref:`global parameters<global options>` such as ``-lang`` or
    ``-log``, initialises the logging layer, which emits startup
    information into log at level 'verbose'. This function makes sure
    that global arguments are applied first, regardless of the order in
    which the arguments were given. ``args`` may be passed as an
    argument, thereby overriding ``sys.argv``.

    >>> local_args = pywikibot.handle_args()  # sys.argv is used
    >>> local_args  # doctest: +SKIP
    []
    >>> local_args = pywikibot.handle_args(['-simulate', '-myoption'])
    >>> local_args  # global options are handled, show the remaining
    ['-myoption']
    >>> for arg in local_args: pass  # do whatever is wanted with local_args

    .. caution::
       Global options might be introduced without warning period. It is
       up to developers to verify that global options do not interfere
       with local script options of private scripts.

    .. tip::
       Avoid using this method in your private scripts and use the
       :mod:`pwb<pywikibot.scripts.wrapper>` wrapper instead. In
       directory mode::

           python pwb.py <global options> <name_of_script> <local options>

       With installed site package::

           pwb <global options> <name_of_script> <local options>

       .. note:: the :mod:`pwb<pywikibot.scripts.wrapper>` wrapper can
          be used even if the `handle_args` method is used within the
          script.

    .. versionchanged:: 5.2
       *-site* global option was added
    .. versionchanged:: 7.1
       *-cosmetic_changes* and *-cc* may be set directly instead of
       toggling the value. Refer :func:`tools.strtobool` for valid values.
    .. versionchanged:: 7.7
       *-config* global option was added.
    .. versionchanged:: 8.0
       Short site value can be given if site code is equal to family
       like ``-site:meta``.
    .. versionchanged:: 8.1
       ``-nolog`` option also discards command.log.

    :param args: Command line arguments. If None,
        :meth:`pywikibot.argvu<userinterfaces._interface_base.ABUIC.argvu>`
        is used which is a copy of ``sys.argv``
    :param do_help: Handle parameter '-help' to show help and invoke sys.exit
    :return: list of arguments not recognised globally
    """
    if pywikibot._sites:
        warnings.warn('Site objects have been created before arguments were '
                      'handled', UserWarning, stacklevel=2)

    # get commandline arguments if necessary
    if not args:
        # it's the version in pywikibot.__init__ that is changed by scripts,
        # not the one in pywikibot.bot.
        args = pywikibot.argvu[1:]

    global global_args
    global_args = args

    # get the name of the module calling this function. This is
    # required because the -help option loads the module's docstring and
    # because the module name will be used for the filename of the log.
    module_name = calledModuleName() or 'terminal-interface'
    non_global_args = []
    username = None
    commandlog = True
    do_help_val: bool | str | None = None if do_help else False
    assert args is not None
    for arg in args:
        option, _, value = arg.partition(':')
        if do_help_val is not False and option == '-help':
            do_help_val = value or True
        # these are handled by config.py
        elif option in ('-config', '-dir'):
            pass
        elif option == '-site':
            if ':' in value:
                config.family, config.mylang = value.split(':')
            else:
                config.family = config.mylang = value
        elif option == '-family':
            config.family = value
        elif option == '-lang':
            config.mylang = value
        elif option == '-user':
            username = value
        elif option in ('-putthrottle', '-pt'):
            config.put_throttle = float(value)
        elif option == '-log':
            if module_name not in config.log:
                config.log.append(module_name)
            if value:
                config.logfilename = value
        elif option == '-nolog':
            commandlog = False
            config.log = []
        elif option in ('-cosmeticchanges', '-cc'):
            config.cosmetic_changes = (strtobool(value) if value
                                       else not config.cosmetic_changes)
            pywikibot.info(f'NOTE: option cosmetic_changes is '
                           f'{config.cosmetic_changes}\n')
        elif option == '-simulate':
            config.simulate = value or True
        #
        #  DEBUG control:
        #
        #    The framework has four layers (by default, others can be added),
        #    each designated by a string --
        #
        #    1.  "comm": the communication layer (http requests, etc.)
        #    2.  "data": the raw data layer (API requests, XML dump parsing)
        #    3.  "wiki": the wiki content representation layer (Page and Site
        #         objects)
        #    4.  "bot": the application layer (user scripts should always
        #         send any debug() messages to this layer)
        #
        #    The "-debug:layer" flag sets the logger for any specified
        #    layer to the DEBUG level, causing it to output extensive debugging
        #    information. Otherwise, the default logging setting is the INFO
        #    level. "-debug" with no layer specified sets _all_ loggers to
        #    DEBUG level.
        #
        #    This method does not check the 'layer' part of the flag for
        #    validity.
        #
        #    If used, "-debug" turns on file logging, regardless of any
        #    other settings.
        #
        elif option == '-debug':
            if module_name not in config.log:
                config.log.append(module_name)
            if value not in config.debug_log:
                config.debug_log.append(value)  # may be empty string
        elif option in ('-verbose', '-v'):
            config.verbose_output += 1
        elif option == '-daemonize':
            redirect_std = value or None
            daemonize.daemonize(redirect_std=redirect_std)
        else:
            # the argument depends on numerical config settings
            # e.g. -maxlag and -step:
            try:
                _arg = option[1:]
                # explicitly check for int (so bool doesn't match)
                if not isinstance(getattr(config, _arg), int):
                    raise TypeError
                setattr(config, _arg, int(value))
            except (ValueError, TypeError, AttributeError):
                # argument not global -> specific bot script will take care
                non_global_args.append(arg)

    if username:
        config.usernames[config.family][config.mylang] = username

    init_handlers()
    if commandlog:
        writeToCommandLogFile()

    if config.verbose_output:
        pywikibot.info('Python ' + sys.version)

    if do_help_val:
        show_help(show_global=do_help_val == 'global')
        sys.exit(0)

    if calledModuleName() != 'generate_user_files':  # T261771
        try:
            pywikibot.Site()
        except (UnknownFamilyError, UnknownSiteError):
            _exception(exc_info=False)
            sys.exit(1)
        if calledModuleName() == 'wrapper':
            pywikibot._sites.clear()

    _debug('handle_args() completed.')
    return non_global_args


def show_help(module_name: str | None = None,
              show_global: bool = False) -> None:
    """Show help for the Bot.

    .. versionchanged:: 4.0
       Renamed from showHelp() to show_help().
    .. versionchanged:: 8.0
       Do not show version changes.
    """
    if not module_name:
        module_name = calledModuleName()
    if not module_name:
        try:
            main = sys.modules['__main__'].main  # type: ignore[attr-defined]
            module_name = main.__module__
            assert module_name is not None
        except NameError:
            module_name = 'no_module'

    try:
        module = import_module(module_name)
    except ModuleNotFoundError:
        if module_name:
            _stdout('Sorry, no help available for ' + module_name)
        _log('show_help:', exc_info=True)
    else:
        assert module.__doc__ is not None
        help_text = re.sub(r'^\.\. version(added|changed)::.+', '',
                           module.__doc__, flags=re.MULTILINE | re.DOTALL)
        if hasattr(module, 'docuReplacements'):
            for key, value in module.docuReplacements.items():
                help_text = help_text.replace(key, value.strip())
        _stdout(help_text)

    if show_global or module_name == 'pwb':
        _stdout(_GLOBAL_HELP.format(module_name))
    else:
        _stdout(_GLOBAL_HELP_NOTE)


def suggest_help(missing_parameters: Sequence[str] | None = None,
                 missing_generator: bool = False,
                 unknown_parameters: Sequence[str] | None = None,
                 exception: Exception | None = None,
                 missing_action: bool = False,
                 additional_text: str = '',
                 missing_dependencies: Sequence[str] | None = None) -> bool:
    """Output error message to use -help with additional text before it.

    :param missing_parameters: A list of parameters which are missing.
    :param missing_generator: Whether a generator is missing.
    :param unknown_parameters: A list of parameters which are unknown.
    :param exception: An exception thrown.
    :param missing_action: Add an entry that no action was defined.
    :param additional_text: Additional text added to the end.
    :param missing_dependencies: A list of dependencies which cannot be
        imported.
    :return: True if an error message was printed, False otherwise
    """
    messages = []
    if exception:
        messages.append(f'An error occurred: "{exception}".')
    if missing_generator:
        messages.append(
            'Unable to execute script because no generator was defined.')
    if missing_parameters:
        messages.append('Missing parameter{s} "{params}".'
                        .format(s='s' if len(missing_parameters) > 1 else '',
                                params='", "'.join(missing_parameters)))
    if missing_action:
        messages.append('No action defined.')
    if unknown_parameters:
        messages.append('Unknown parameter{s} "{params}".'
                        .format(s='s' if len(unknown_parameters) > 1 else '',
                                params='", "'.join(unknown_parameters)))
    if missing_dependencies:
        messages.append('Missing dependenc{s} "{deps}".'
                        .format(
                            s='ies' if len(missing_dependencies) > 1 else 'y',
                            deps='", "'.join(missing_dependencies)))
    if additional_text:
        messages.append(additional_text.strip())
    if messages:
        messages.append('Use -help for further information.')
        _error('\n'.join(messages))
        return True
    return False


def writeToCommandLogFile() -> None:
    """Save name of the called module along with all params to logfile.

    This can be used by user later to track errors or report bugs.
    """
    modname = calledModuleName()
    # put quotation marks around all parameters
    args = [modname] + [f'"{s}"' for s in pywikibot.argvu[1:]]
    iso_date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())

    command_log = Path(config.datafilepath('logs', 'commands.log'))
    mode = 'a' if command_log.exists() else 'w'
    with command_log.open(mode, encoding='utf-8') as command_log_file:
        command_log_file.write('{} r{} Python {} '
                               .format(iso_date,
                                       version.getversiondict()['rev'],
                                       sys.version.split()[0]))
        command_log_file.write(' '.join(args) + os.linesep)


def open_webbrowser(page: pywikibot.page.BasePage) -> None:
    """Open the web browser displaying the page and wait for input.

    .. versionchanged:: 10.1
       No longer follow a redirect.
    """
    webbrowser.open(f'{page.full_url()}?redirect=no')
    i18n.input('pywikibot-enter-finished-browser')


class _OptionDict(Dict[str, Any]):

    """The option dict which holds the options of OptionHandler.

    .. versionadded:: 4.1
    """

    def __init__(self, classname: str, options: dict[str, Any]) -> None:
        self._classname = classname
        super().__init__(options)

    def __missing__(self, key: str) -> None:
        raise Error(f"'{key}' is not a valid option for {self._classname}.")

    def __getattr__(self, name: str) -> Any:
        """Get item from dict."""
        return self.__getitem__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set item or attribute."""
        if name != '_classname':
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)


class OptionHandler:

    """Class to get and set options.

    How to use options of OptionHandler and its BaseBot subclasses:
    First define an available_options class attribute for your own
    option handler to define all available options:

    >>> default_options = {'foo': 'bar', 'bar': 42, 'baz': False}
    >>> class MyHandler(OptionHandler): available_options = default_options

    Or you may update the predefined setting in the class initializer.
    BaseBot predefines a 'always' options and sets it to False:

    self.available_options.update(always=True, another_option='Yes')

    Now you can instantiate an OptionHandler or BaseBot class passing
    options other than default values:

    >>> bot = MyHandler(baz=True)

    You can access bot options either as keyword item or attribute:

    >>> bot.opt.foo
    'bar'
    >>> bot.opt['bar']
    42
    >>> bot.opt.baz  # default was overridden
    True

    You can set the options in the same way:

    >>> bot.opt.bar = 4711
    >>> bot.opt['baz'] = None
    >>>

    Or you can use the option as a dict:

    >>> 'Option opt.{foo} is {bar}'.format_map(bot.opt)
    'Option opt.bar is 4711'

    .. warning:: You must not access bot options as an attribute if the
       keyword is a :python:`dict method<library/stdtypes.html#dict.clear>`.
    """

    available_options: dict[str, Any] = {}
    """Handler configuration attribute.

    Only the keys of the dict can be passed as `__init__` options.
    The values are the default values. Overwrite this in subclasses!
    """

    def __init__(self, **kwargs: Any) -> None:
        """Only accept options defined in available_options.

        :param kwargs: bot options
        """
        self.set_options(**kwargs)

    def set_options(self, **options: Any) -> None:
        """Set the instance options."""
        valid_options = set(self.available_options)
        received_options = set(options)

        # self.opt contains all available options including defaults
        self.opt = _OptionDict(self.__class__.__name__, self.available_options)
        # update the options overridden from defaults
        self.opt.update((opt, options[opt])
                        for opt in received_options & valid_options)
        for opt in received_options - valid_options:
            _warning(f'{opt} is not a valid option. It was ignored.')


class BaseBot(OptionHandler):

    """Generic Bot to be subclassed.

    Only accepts `generator` and options defined in
    :attr:`available_options`.

    This class provides a :meth:`run` method for basic processing of a
    generator one page at a time.

    If the subclass places a page generator in :attr:`generator`, Bot
    will process each page in the generator, invoking the method
    :meth:`treat` which must then be implemented by subclasses.

    Each item processed by :meth:`treat` must be a :class:`page.BasePage`
    type. Use :meth:`init_page` to upcast the type. To enable other
    types, set :attr:`BaseBot.treat_page_type` to an appropriate type;
    your bot should derive from :class:`BaseBot` in that case and handle
    site properties.

    If the subclass does not set a generator, or does not override
    :meth:`treat` or :meth:`run`, `NotImplementedError` is raised.

    For bot options handling refer :class:`OptionHandler` class above.

    .. versionchanged:: 7.0
       A :attr:`counter` instance variable is provided.
    """

    use_disambigs: bool | None = None
    """Attribute to determine whether to use disambiguation pages. Set it to
    True to use disambigs only, set it to False to skip disambigs. If None both
    are processed.

    .. versionadded:: 7.2
    """

    use_redirects: bool | None = None
    """Attribute to determine whether to use redirect pages. Set it to
    True to use redirects only, set it to False to skip redirects. If
    None both are processed. For example to create a RedirectBot you may
    define:

    .. code-block:: python

       class MyRedirectBot(ExistingPageBot):

           '''Bot who only works on existing redirects.'''

           use_redirects = True

    .. versionadded:: 7.2
    """

    available_options = {
        'always': False,  # By default ask for confirmation when putting a page
    }

    update_options: dict[str, Any] = {}
    """`update_options` can be used to update :attr:`available_options`; do not
    use it if the bot class is to be derived but use
    `self.available_options.update(<dict>)` initializer in such case.

    .. versionadded:: 6.4
    """

    _current_page: pywikibot.page.BasePage | None = None

    def __init__(self, **kwargs: Any) -> None:
        """Initializer.

        :param kwargs: bot options
        :keyword generator: a :attr:`generator` processed by :meth:`run`
            method
        """
        if 'generator' in kwargs:
            if hasattr(self, 'generator'):
                warnings.warn(f'{type(self).__name__} has a generator'
                              ' already. Ignoring argument.', stacklevel=2)
            else:
                self.generator: Iterable = kwargs.pop('generator')

        self.available_options.update(self.update_options)
        super().__init__(**kwargs)

        self.counter: Counter = Counter()
        """Instance variable which holds counters. The default counters
        are 'read', 'write' and 'skip'. All of them are printed within
        :meth:`exit`. You can use your own counters like::

            self.counter['delete'] += 1

        .. versionadded:: 7.0
        .. versionchanged:: 7.3
           Your additional counters are also printed during :meth:`exit`
        """

        self.generator_completed: bool = False
        """Instance attribute which is True if the :attr:`generator` is
        completed.

        It gives False if the the generator processing in :meth:`run` is
        either interrupted by ``KeyboardInterrupt`` or exited by
        :exc:`QuitKeyboardInterrupt` while closing the generator i.e.
        :code:`self.generator.close()` keeps the value True.

        To check for an empty generator you may use::

            if self.generator_completed and not self.counter['read']:
                print('generator was emtpty')

        .. note:: An empty generator returns True.
        .. versionadded:: 3.0
        .. versionchanged:: 7.4
           renamed to `generator_completed` to become a public attribute.
        """

        self.treat_page_type: Any = pywikibot.page.BasePage
        """Instance variable to hold the default page type used by :meth:`run`.

        .. versionadded:: 6.1
        """

    @property
    def current_page(self) -> pywikibot.page.BasePage:
        """Return the current working page as a property."""
        assert self._current_page is not None
        return self._current_page

    @current_page.setter
    def current_page(self, page: pywikibot.page.BasePage) -> None:
        """Set the current working page as a property.

        When the value is actually changed, the page title is printed to
        the standard output (highlighted in purple) and logged with a
        VERBOSE level.

        This also prevents the same title from being printed twice.

        :param page: the working page
        """
        if page != self._current_page:
            self._current_page = page
            msg = f'Working on {page.title()!r}'
            if config.colorized_output:
                _log(msg)
                _stdout(
                    f'\n\n>>> <<lightpurple>>{page.title()}<<default>> <<<')
            else:
                _stdout(msg)

    def user_confirm(self, question: str) -> bool:
        """Obtain user response if bot option 'always' not enabled."""
        if self.opt.always:
            return True

        choice = pywikibot.input_choice(question,
                                        [('Yes', 'y'),
                                         ('No', 'N'),
                                         ('All', 'a'),
                                         ('Quit', 'q')],
                                        default='N',
                                        automatic_quit=False)

        if choice == 'n':
            return False

        if choice == 'q':
            self.quit()

        if choice == 'a':
            # Remember the choice
            self.opt.always = True

        return True

    def userPut(self, page: pywikibot.page.BasePage, oldtext: str,
                newtext: str, **kwargs: Any) -> bool:
        """Save a new revision of a page, with user confirmation as required.

        Print differences, ask user for confirmation, and puts the page
        if needed.

        Option used:

        * 'always'

        :keyword asynchronous: passed to page.save
        :keyword summary: passed to page.save
        :keyword show_diff: show changes between oldtext and newtext (enabled)
        :keyword ignore_save_related_errors: report and ignore (disabled)
        :keyword ignore_server_errors: report and ignore (disabled)
        :return: whether the page was saved successfully
        """
        if oldtext.rstrip() == newtext.rstrip():
            pywikibot.info(f'No changes were needed on {page}')
            return False

        self.current_page = page

        show_diff = kwargs.pop('show_diff', True)
        if show_diff:
            pywikibot.showDiff(oldtext, newtext)

        if 'summary' in kwargs:
            pywikibot.info(f"Edit summary: {kwargs['summary']}")

        page.text = newtext
        return self._save_page(page, page.save, **kwargs)

    def _save_page(self, page: pywikibot.page.BasePage,
                   func: Callable[..., Any], *args: Any,
                   **kwargs: Any) -> bool:
        """Helper function to handle page save-related option error handling.

        .. note:: Do no use it directly. Use :meth:`userPut` instead.

        :param page: currently edited page
        :param func: the function to call
        :param args: passed to the function
        :param kwargs: passed to the function
        :keyword ignore_server_errors: if True, server errors will be reported
            and ignored (default: False)
        :kwtype ignore_server_errors: bool
        :keyword ignore_save_related_errors: if True, errors related to
            page save will be reported and ignored (default: False)
        :kwtype ignore_save_related_errors: bool
        :return: whether the page was saved successfully

        :meta public:
        """
        if not self.user_confirm('Do you want to accept these changes?'):
            return False

        if 'asynchronous' not in kwargs and self.opt.always:
            kwargs['asynchronous'] = True

        ignore_save_related_errors = kwargs.pop('ignore_save_related_errors',
                                                False)
        ignore_server_errors = kwargs.pop('ignore_server_errors', False)

        try:
            func(*args, **kwargs)
            self.counter['write'] += 1
        except PageSaveRelatedError as e:
            if not ignore_save_related_errors:
                raise
            if isinstance(e, EditConflictError):
                pywikibot.info(
                    f'Skipping {page.title()} because of edit conflict')
            elif isinstance(e, SpamblacklistError):
                pywikibot.info(f'Cannot change {page.title()} because of '
                               f'blacklist entry {e.url}')
            elif isinstance(e, LockedPageError):
                pywikibot.info(f'Skipping {page.title()} (locked page)')
            else:
                _error(f'Skipping {page} because of a save related error: {e}')
        except ServerError as e:
            if not ignore_server_errors:
                raise
            _error(
                f'Server Error while processing {page.title()}: {e}')
        else:
            return True
        return False

    def quit(self) -> None:
        """Cleanup and quit processing."""
        raise QuitKeyboardInterrupt

    def exit(self) -> None:
        """Cleanup and exit processing.

        Invoked when :meth:`run` is finished. Waits for pending threads,
        prints counter statistics and informs whether the script
        terminated gracefully or was halted by exception.

        .. note:: Do not overwrite it by subclasses; :meth:`teardown`
           should be used instead.

        .. versionchanged:: 7.3
           Statistics are printed for all entries in :attr:`counter`
        .. versionchanged:: 9.0
           Print execution time with days, hours, minutes and seconds.
        """
        self.teardown()
        if hasattr(self, '_start_ts'):
            read_delta = pywikibot.Timestamp.now() - self._start_ts
            read_seconds = int(read_delta.total_seconds())

        # wait until pending threads finished but don't close the queue
        pywikibot.stopme()

        pywikibot.info()
        for op, count in self.counter.items():
            pywikibot.info(f"{count} {op} operation{'s' if count > 1 else ''}")

        if hasattr(self, '_start_ts'):
            write_delta = pywikibot.Timestamp.now() - self._start_ts
            write_seconds = int(write_delta.total_seconds())
            hours, remainder = divmod(write_delta.seconds, 3600)
            parts = write_delta.days, hours, *divmod(remainder, 60)
            units = ('days', 'hours', 'minutes', 'seconds')
            used = ', '.join(f'{value} {unit}'
                             for value, unit in zip(parts, units) if value)
            pywikibot.info('Execution time: ' + used)

            if self.counter['read']:
                pywikibot.info('Read operation time: {:.1f} seconds'
                               .format(read_seconds / self.counter['read']))

            for op, count in self.counter.items():
                if not count or op == 'read':
                    continue
                pywikibot.info(f'{op.capitalize()} operation time: '
                               f'{write_seconds / count:.1f} seconds')

        # exc_info contains exception from self.run() while terminating
        exc_info = sys.exc_info()
        pywikibot.info('Script terminated ', newline=False)
        if exc_info[0] is None or exc_info[0] is KeyboardInterrupt:
            pywikibot.info('successfully.')
        else:
            pywikibot.info('by exception:\n')
            _exception(exc_info=False)

    def init_page(self, item: Any) -> pywikibot.page.BasePage:
        """Initialize a generator item before treating.

        Ensure that the result of `init_page` is always a
        pywikibot.Page object or any other type given by the
        :attr:`treat_page_type` even when the generator returns
        something else.

        Also used to set the arrange the current site. This is called
        before :meth:`skip_page` and :meth:`treat`.

        :param item: any item from :attr:`generator`
        :return: return the page object to be processed further
        """
        return item

    def skip_page(self, page: pywikibot.page.BasePage) -> bool:
        """Return whether treat should be skipped for the page.

        .. versionadded:: 3.0

        .. versionchanged:: 7.2
           use :attr:`use_redirects` to handle redirects,
           use :attr:`use_disambigs` to handle disambigs

        :param page: Page object to be processed
        """
        if isinstance(self.use_redirects, bool) \
           and page.isRedirectPage() is not self.use_redirects:
            _warning(
                'Page {page} on {page.site} is skipped because it is {not_}'
                'a redirect'
                .format(page=page, not_='not ' if self.use_redirects else ''))
            return True

        if isinstance(self.use_disambigs, bool) \
           and page.isDisambig() is not self.use_disambigs:
            _warning(
                'Page {page} on {page.site} is skipped because it is {not_}'
                'a disambig'
                .format(page=page, not_='not ' if self.use_disambigs else ''))
            return True

        return False

    def treat(self, page: Any) -> None:
        """Process one page (abstract method).

        :param page: Object to be processed, usually a
            :class:`page.BasePage`. For other page types the
            :attr:`treat_page_type` must be set.
        """
        raise NotImplementedError(
            f'Method {type(self).__name__}.treat() not implemented.')

    def setup(self) -> None:
        """Some initial setup before :meth:`run` operation starts.

        This can be used for reading huge parts from life wiki or file
        operation which is more than just initialize the instance.
        Invoked by :meth:`run` before running through :attr:`generator`
        loop.

        .. versionadded:: 3.0
        """

    def teardown(self) -> None:
        """Some cleanups after :meth:`run` operation. Invoked by :meth:`exit`.

        .. versionadded:: 3.0
        """

    def run(self) -> None:
        """Process all pages in generator.

        Call :meth:`setup`, check for a valid ``Iterable`` type in
        :attr:`generator`, upcast it to a ``Generator`` type if
        necessary, process every generator`s item as follows:

        For each item call :meth:`init_page`, check whether the result
        is a :attr:`treat_page_type` type, call :meth:`skip_page` to
        determine whether to skip the current page. Otherwise call
        :meth:`treat` for each item.

        This method also adjust ``read`` and ``skip`` :attr:`counter`,
        and finally it calls :meth:`exit` when leaving the method. In
        short this method is implemented similar to this:

        .. code-block:: python

           def run(self) -> None:
               '''Process all pages in generator.'''
               self.setup()

               if not hasattr(self, 'generator'):
                   raise NotImplementedError('"generator" not set.')

               if self.generator is None;
                   print('No generator was defined')

               try:
                   for item in self.generator:
                       page = self.init_page(item)

                   if self.skip_page(page):
                       continue

                   self.treat(page)

               except(QuitKeyboardInterrupt, KeyboardInterrupt):
                   print('User canceled bot run.')

               finally:
                   self.exit()

        .. versionchanged:: 3.0
           ``skip`` counter was added.; call :meth:`setup` first.
        .. versionchanged:: 6.0
           upcast :attr:`generator` to a ``Generator`` type to enable
           ``generator.close()`` method.
        .. versionchanged:: 6.1
           Objects from :attr:`generator` may be different from
           :class:`pywikibot.Page` but the type must be registered in
           :attr:`treat_page_type`.
        .. versionchanged:: 9.2
           leave method gracefully if :attr:`generator` is None using
           :func:`suggest_help` function.

        :raise AssertionError: "page" is not a pywikibot.page.BasePage
            object
        :raise KeyboardInterrupt: KeyboardInterrupt occurred while
            :attr:`config.verbose_output` was set
        :raise NotImplementedError: :attr:`generator` is not set
        :raise TypeError: invalid generator type or page is not a
            :attr:`treat_page_type`
        """
        self._start_ts = pywikibot.Timestamp.now()
        self.setup()

        if not hasattr(self, 'generator'):
            raise NotImplementedError(
                f'Variable {type(self).__name__}.generator not set.')

        if suggest_help(missing_generator=self.generator is None):
            return

        if not isinstance(self.generator, Generator):
            gen_type = type(self.generator).__name__
            _debug(f'wrapping {gen_type} type to a Generator type')
            try:
                # to provide generator.close() method
                self.generator = (item for item in self.generator)
            except TypeError:
                raise TypeError(f'Invalid type {gen_type} for generator')

        try:
            for item in self.generator:
                # preprocessing of the page
                page = self.init_page(item)

                # validate page type
                if not isinstance(page, self.treat_page_type):
                    raise TypeError(f'"page" is not a {self.treat_page_type!r}'
                                    f' object but {type(page).__name__}.')

                if self.skip_page(page):
                    self.counter['skip'] += 1
                    continue

                # Process the page
                self.counter['read'] += 1
                self.treat(page)

            self.generator_completed = True
        except QuitKeyboardInterrupt:
            pywikibot.info(f'\nUser quit {type(self).__name__} bot run...')
        except KeyboardInterrupt:
            if config.verbose_output:
                raise

            pywikibot.info(
                f'\nKeyboardInterrupt during {type(self).__name__} bot run...')
        finally:
            self.exit()


# TODO: Deprecate Bot class as self.site may be the site of the page or may be
# a site previously defined
class Bot(BaseBot):

    """Generic bot subclass for multiple sites.

    If possible the MultipleSitesBot or SingleSiteBot classes should be
    used instead which specifically handle multiple or single sites.
    """

    def __init__(self, site: BaseSite | None = None,
                 **kwargs: Any) -> None:
        """Create a Bot instance and initialize cached sites."""
        # TODO: add warning if site is specified and generator
        # contains pages from a different site.
        self._site = site
        self._sites = set([self._site] if self._site else [])

        super().__init__(**kwargs)

    @property
    def site(self) -> BaseSite | None:
        """Get the current site."""
        if not self._site:
            _warning('Bot.site was not set before being retrieved.')
            self.site = pywikibot.Site()
            _warning(f'Using the default site: {self.site}')
        assert self._site is not None
        return self._site

    @site.setter
    def site(self, site: BaseSite | None) -> None:
        """Set the Site that the bot is using.

        When Bot.run() is managing the generator and site property, this
        is set each time a page is on a site different from the previous
        page.
        """
        if not site:
            self._site = None
            return

        if site not in self._sites:
            _log(f'LOADING SITE {site} VERSION: {site.mw_version}')

            self._sites.add(site)
            if len(self._sites) == 2:
                _log(f'{self.__class__.__name__} uses multiple sites')

        if self._site and self._site != site:
            _log(f'{type(self).__name__}: changing site from {self._site} '
                 f'to {site}')

        self._site = site

    def run(self) -> None:
        """Check if it automatically updates the site before run."""
        # This check is to remove the possibility that the superclass changing
        # self.site causes bugs in subclasses.
        # If the subclass has set self.site before run(), it may be that the
        # bot processes pages on sites other than self.site, and therefore
        # this method can't alter self.site. To use this functionality, don't
        # set self.site in __init__, and use page.site in treat().
        self._auto_update_site = not self._site
        if not self._auto_update_site:
            _warning(
                f'{type(self).__name__}.__init__ set the Bot.site property;'
                ' this is only needed when the Bot accesses many sites.'
            )
        else:
            _log(f'Bot is managing the {type(self).__name__}.site property in'
                 ' run()')
        super().run()

    def init_page(self, item: Any) -> pywikibot.page.BasePage:
        """Update site before calling treat."""
        # When in auto update mode, set the site when it changes,
        # so subclasses can hook onto changes to site.
        page = super().init_page(item)
        if (self._auto_update_site
                and (not self._site or page.site != self.site)):
            self.site = page.site
        return page


class SingleSiteBot(BaseBot):

    """A bot only working on one site and ignoring the others.

    If no site is given from the start it'll use the first page's site.
    Any page after the site has been defined and is not on the defined
    site will be ignored.
    """

    def __init__(self,
                 site: BaseSite | bool | None = True,
                 **kwargs: Any) -> None:
        """Create a SingleSiteBot instance.

        :param site: If True it'll be set to the configured site using
            pywikibot.Site.
        """
        if site is True:
            self._site: BaseSite | None = pywikibot.Site()
        elif site is False:
            raise ValueError("'site' must be a site, True, or None")
        else:
            self._site = site
        super().__init__(**kwargs)

    @property
    def site(self) -> BaseSite:
        """Site that the bot is using."""
        if not self._site:
            raise ValueError('The site has not been defined yet.')
        return self._site

    @site.setter
    def site(self, value: BaseSite | None) -> None:
        """Set the current site but warns if different."""
        if self._site:
            # Warn in any case where the site is (probably) changed after
            # setting it the first time. The appropriate variant is not to use
            # self.site at all or define it once and never change it again
            if self._site == value:
                _warning('Defined site without changing it.')
            else:
                _warning(f'Changed the site from "{self._site}" to "{value}"')
        self._site = value

    def init_page(self, item: Any) -> pywikibot.page.BasePage:
        """Set site if not defined."""
        page = super().init_page(item)
        if not self._site:
            self.site = page.site
        return page

    def skip_page(self, page: pywikibot.page.BasePage) -> bool:
        """Skip page if it is not on the defined site."""
        if page.site != self.site:
            _warning(fill(f'Skipped {page} due to: '
                          f'The bot is on site "{self.site}" but the page on '
                          f'site "{page.site}"'))
            return True
        return super().skip_page(page)


class MultipleSitesBot(BaseBot):

    """A bot class working on multiple sites.

    The bot should accommodate for that case and not store site specific
    information on only one site.

    .. versionchanged:: 6.2
       Site attribute has been dropped.
    """


class ConfigParserBot(BaseBot):

    """A bot class that can read options from scripts.ini file.

    All options must be predefined in available_options dictionary. The type
    of these options is responsible for the correct interpretation of the
    options type given by the .ini file. They can be interpreted as bool,
    int, float or str (default). The settings file may be like::

        [add_text]
        # edit summary for the bot.
        summary = Bot: Aggiungo template Categorizzare

        [commonscat] ; commonscat options
        always: true

    The option values are interpreted in this order:

    1. `available_options` default setting
    2. `script.ini options` settings
    3. command line arguments

    .. versionadded:: 3.0
    """

    INI = 'scripts.ini'

    def set_options(self, **kwargs: Any) -> None:
        """Read settings from scripts.ini file."""
        conf = configparser.ConfigParser(inline_comment_prefixes=[';'])
        section = calledModuleName()

        if (conf.read(self.INI) == [self.INI] and conf.has_section(section)):
            pywikibot.info(f'Reading settings from {self.INI} file.')
            options = {}
            for option, value in self.available_options.items():
                if not conf.has_option(section, option):
                    continue
                # use a convenience parser method, default to get()
                default = conf.get
                value_type = type(value).__name__
                if value_type == 'bool':
                    method = conf.getboolean
                else:
                    method = getattr(conf, 'get' + value_type, default)
                options[option] = method(section, option)
            for opt in set(conf.options(section)) - set(options):
                _warning(f'"{opt}" is not a valid option. It was ignored.')
            options.update(kwargs)
        else:
            options = kwargs

        super().set_options(**options)


class CurrentPageBot(BaseBot):

    """A bot which automatically sets 'current_page' on each treat().

    This class should be always used together with either the
    MultipleSitesBot or SingleSiteBot class as there is no site
    management in this class.
    """

    ignore_save_related_errors = True
    ignore_server_errors = False

    def treat_page(self) -> None:
        """Process one page (Abstract method)."""
        raise NotImplementedError(
            f'Method {type(self).__name__}.treat_page() not implemented.')

    def treat(self, page: pywikibot.page.BasePage) -> None:
        """Set page to current page and treat that page."""
        self.current_page = page
        self.treat_page()

    def put_current(
        self,
        new_text: str,
        ignore_save_related_errors: bool | None = None,
        ignore_server_errors: bool | None = None,
        **kwargs: Any
    ) -> bool:
        """Call :py:obj:`Bot.userPut` but use the current page.

        It compares the new_text to the current page text.

        :param new_text: The new text
        :param ignore_save_related_errors: Ignore save related errors
            and automatically print a message. If None uses this
            instances default.
        :param ignore_server_errors: Ignore server errors and
            automatically print a message. If None uses this instances
            default.
        :param kwargs: Additional parameters directly given to
            :meth:`BaseBot.userPut`.
        :return: whether the page was saved successfully
        """
        if ignore_save_related_errors is None:
            ignore_save_related_errors = self.ignore_save_related_errors
        if ignore_server_errors is None:
            ignore_server_errors = self.ignore_server_errors
        return self.userPut(
            self.current_page,
            self.current_page.text,
            new_text,
            ignore_save_related_errors=ignore_save_related_errors,
            ignore_server_errors=ignore_server_errors,
            **kwargs
        )


class AutomaticTWSummaryBot(CurrentPageBot):

    """A class which automatically defines ``summary`` for ``put_current``.

    The class must defined a ``summary_key`` string which contains the
    i18n key for :py:obj:`i18n.twtranslate`. It can also
    override the ``summary_parameters`` property to specify any
    parameters for the translated message.
    """

    #: Must be defined in subclasses.
    summary_key: str | None = None

    @property
    def summary_parameters(self) -> dict[str, str]:
        """A dictionary of all parameters for i18n."""
        if hasattr(self, '_summary_parameters'):
            return self._summary_parameters
        return {}

    @summary_parameters.setter
    def summary_parameters(self, value: dict[str, str]) -> None:
        """Set the i18n dictionary."""
        if not isinstance(value, dict):
            raise TypeError(
                f'"value" must be a dict but {type(value).__name__} was found.'
            )
        self._summary_parameters = value

    @summary_parameters.deleter
    def summary_parameters(self) -> None:
        """Delete the i18n dictionary."""
        del self._summary_parameters

    def put_current(self, *args: Any, **kwargs: Any) -> bool:
        """Defining a summary if not already defined and then call original.

        For parameters see :meth:`CurrentPageBot.put_current`

        .. versionchanged:: 10.6
           return whether the page was saved successfully
        """
        if not kwargs.get('summary'):
            if self.summary_key is None:
                raise ValueError('The summary_key must be set.')
            summary = i18n.twtranslate(self.current_page.site,
                                       self.summary_key,
                                       self.summary_parameters)
            _log(f'Use automatic summary message "{summary}"')
            kwargs['summary'] = summary

        return super().put_current(*args, **kwargs)


class ExistingPageBot(CurrentPageBot):

    """A CurrentPageBot class which only treats existing pages."""

    def skip_page(self, page: pywikibot.page.BasePage) -> bool:
        """Treat page if it exists and handle NoPageError.

        .. warning:: If subclassed, call `super().skip_page()` first to
           ensure that non existent pages are filtered before other
           calls are made
        """
        if not page.exists():
            _warning(f'Page {page} does not exist on {page.site}.')
            return True
        return super().skip_page(page)


class FollowRedirectPageBot(CurrentPageBot):

    """A CurrentPageBot class which follows the redirect."""

    def treat(self, page: pywikibot.page.BasePage) -> None:
        """Treat target if page is redirect and the page otherwise."""
        if page.isRedirectPage():
            page = page.getRedirectTarget()
        super().treat(page)


class CreatingPageBot(CurrentPageBot):

    """A CurrentPageBot class which only treats nonexistent pages."""

    def skip_page(self, page: pywikibot.page.BasePage) -> bool:
        """Treat page if doesn't exist."""
        if page.exists():
            _warning(f'Page {page} does already exist on {page.site}.')
            return True
        return super().skip_page(page)


class WikidataBot(Bot, ExistingPageBot):

    """Generic Wikidata Bot to be subclassed.

    Source claims (P143) can be created for specific sites

    :cvar use_from_page: If True (default) it will apply
        ItemPage.fromPage for every item. If False it assumes that the
        pages are actually already ItemPage (page in treat_page_and_item
        will be None). If None it'll use ItemPage.fromPage when the page
        is not in the site's item namespace.
    :type use_from_page: bool, None
    :cvar treat_missing_item: Whether pages without items should be
        treated. Note that this is checked after create_missing_item.
    :type treat_missing_item: bool
    :ivar create_missing_item: If True, new items will be created if the
        current page doesn't have one. Subclasses should override this
        in the initializer with a bool value or using self.opt
        attribute.
    :type create_missing_item: bool
    """

    use_from_page = True
    treat_missing_item = False

    def __init__(self, **kwargs: Any) -> None:
        """Initializer of the WikidataBot."""
        self.create_missing_item = False
        super().__init__(**kwargs)
        self.site = pywikibot.Site()
        self.repo = self.site.data_repository()
        if self.repo is None:
            raise WikiBaseError(
                f'{self.site} is not connected to a data repository')

    def cacheSources(self) -> None:
        """Fetch the sources from the list on Wikidata.

        It is stored internally and reused by getSource()
        """
        page = pywikibot.Page(self.repo, 'List of wikis/python', ns=4)
        self.source_values = json.loads(page.get())
        for family_code, family in self.source_values.items():
            for source_lang in family:
                self.source_values[
                    family_code][source_lang] = pywikibot.ItemPage(
                        self.repo, family[source_lang])

    def get_property_by_name(self, property_name: str) -> str:
        """Find given property and return its ID.

        Method first uses site.search() and if the property isn't found,
        then asks user to provide the property ID.

        :param property_name: property to find
        """
        ns = self.repo.property_namespace
        for page in self.repo.search(property_name, total=1, namespaces=ns):
            prop = pywikibot.PropertyPage(self.repo, page.title())
            pywikibot.info(
                f'Assuming that {property_name} property is {prop.id}.')
            return prop.id

        return pywikibot.input(
            f'Property {property_name} was not found. Please enter the '
            f'property ID (e.g. P123) of it:').upper()

    def user_edit_entity(self, entity: pywikibot.page.WikibasePage,
                         data: dict[str, str] | None = None,
                         ignore_save_related_errors: bool | None = None,
                         ignore_server_errors: bool | None = None,
                         **kwargs: Any) -> bool:
        """Edit entity with data provided, with user confirmation as required.

        :param entity: page to be edited
        :param data: data to be saved, or None if the diff should be
            created automatically
        :param ignore_save_related_errors: Ignore save related errors
            and automatically print a message. If None uses this
            instances default.
        :param ignore_server_errors: Ignore server errors and
            automatically print a message. If None uses this instances
            default.
        :keyword summary: revision comment, passed to
            ItemPage.editEntity
        :keyword show_diff: show changes between oldtext and newtext
            (default: True)
        :return: whether the item was saved successfully
        """
        if ignore_save_related_errors is None:
            ignore_save_related_errors = self.ignore_save_related_errors
        if ignore_server_errors is None:
            ignore_server_errors = self.ignore_server_errors
        show_diff = kwargs.pop('show_diff', True)
        if show_diff:
            if data is None:
                diff = entity.toJSON(diffto=getattr(entity, '_content', None))
            else:
                diff = entity._normalizeData(data)
            pywikibot.info(json.dumps(diff, indent=4, sort_keys=True))

        if 'summary' in kwargs:
            pywikibot.info(f"Change summary: {kwargs['summary']}")

        # TODO PageSaveRelatedErrors should be actually raised in editEntity
        # (bug T86083)
        return self._save_page(
            entity, entity.editEntity, data,
            ignore_save_related_errors=ignore_save_related_errors,
            ignore_server_errors=ignore_server_errors, **kwargs)

    def user_add_claim(self, item: pywikibot.page.ItemPage,
                       claim: pywikibot.page.Claim,
                       source: BaseSite | None = None,
                       bot: bool = True, **kwargs: Any) -> bool:
        """Add a claim to an item, with user confirmation as required.

        :param item: page to be edited
        :param claim: claim to be saved
        :param source: site where the claim comes from
        :param bot: whether to flag as bot (if possible)
        :keyword ignore_server_errors: if True, server errors will be reported
          and ignored (default: False)
        :keyword ignore_save_related_errors: if True, errors related to
          page save will be reported and ignored (default: False)
        :return: whether the item was saved successfully

        .. note:: calling this method sets the current_page property
           to the item which changes the site property

        .. note:: calling this method with the 'source' argument modifies
           the provided claim object in place
        """
        self.current_page = item

        if source:
            sourceclaim = self.getSource(source)
            if sourceclaim:
                claim.addSource(sourceclaim)

        pywikibot.info(f'Adding {claim.getID()} --> {claim.getTarget()}')
        return self._save_page(item, item.addClaim, claim, bot=bot, **kwargs)

    def getSource(self, site: BaseSite) -> pywikibot.page.Claim | None:
        """Create a Claim usable as a source for Wikibase statements.

        :param site: site that is the source of assertions.
        :return: pywikibot.Claim or None
        """
        source = None
        item = i18n.translate(site, self.source_values)
        if item:
            source = pywikibot.Claim(self.repo, 'P143')
            source.setTarget(item)
        return source

    def user_add_claim_unless_exists(
            self, item: pywikibot.page.ItemPage,
            claim: pywikibot.page.Claim,
            exists_arg: Container = '',
            source: BaseSite | None = None,
            logger_callback: Callable[[str], Any] = pwb_logging.log,
            **kwargs: Any) -> bool:
        """Decorator of :py:obj:`user_add_claim`.

        Before adding a new claim, it checks if we can add it, using provided
        filters.

        .. seealso:: documentation of :py:obj:`claimit.py<scripts.claimit>`

        :param exists_arg: pattern for merging existing claims with new ones
        :param logger_callback: function logging the output of the method
        :return: whether the claim could be added

        .. note:: calling this method may change the current_page property
           to the item which will also change the site property

        .. note:: calling this method with the 'source' argument modifies
           the provided claim object in place
        """
        # This code is somewhat duplicate to user_add_claim but
        # unfortunately we need the source claim here, too.
        sourceclaim = self.getSource(source) if source else None

        # Existing claims on page of same property
        claims = item.get().get('claims')
        assert claims is not None

        claim_id = claim.getID()
        for existing in claims.get(claim_id, []):
            # If claim with same property already exists...
            if 'p' not in exists_arg:
                logger_callback(
                    f'Skipping {claim_id} because claim with same property'
                    ' already exists')
                _log('Use -exists:p option to override this behavior')
                break

            if not existing.target_equals(claim.getTarget()):
                continue

            # If some attribute of the claim being added
            # matches some attribute in an existing claim of
            # the same property, skip the claim, unless the
            # 'exists' argument overrides it.
            if 't' not in exists_arg:
                logger_callback(
                    f'Skipping {claim_id} because claim with same target'
                    ' already exists')
                _log("Append 't' to -exists argument to override this "
                     'behavior')
                break

            if 'q' not in exists_arg and not existing.qualifiers:
                logger_callback(
                    f'Skipping {claim_id} because claim without qualifiers'
                    ' already exists')
                _log("Append 'q' to -exists argument to override this "
                     'behavior')
                break

            if ('s' not in exists_arg or not sourceclaim) \
               and not existing.sources:
                logger_callback(
                    f'Skipping {claim_id} because claim without source already'
                    ' exists')
                _log("Append 's' to -exists argument to override this "
                     'behavior')
                break

            # FIXME: the user may provide a better source, but we only
            # assume it's the default one
            if ('s' not in exists_arg and sourceclaim
                and any(sourceclaim.getID() in ref
                        and all(snak.target_equals(sourceclaim.getTarget())
                                for snak in ref[sourceclaim.getID()])
                        for ref in existing.sources)):
                logger_callback(
                    f'Skipping {claim_id} because claim with the same source'
                    ' already exists')
                _log("Append 's' to -exists argument to override this "
                     'behavior')
                break
        else:
            return self.user_add_claim(item, claim, source, **kwargs)

        return False

    def create_item_for_page(self, page: pywikibot.page.BasePage,
                             data: dict[str, Any] | None = None,
                             summary: str | None = None,
                             **kwargs: Any
                             ) -> pywikibot.page.ItemPage | None:
        """Create an ItemPage with the provided page as the sitelink.

        :param page: the page for which the item will be created
        :param data: additional data to be included in the new item
            (optional). Note that data created from the page have higher
            priority.
        :param summary: optional edit summary to replace the default one
        :return: pywikibot.ItemPage or None
        """
        if not summary:
            summary = ('Bot: New item with sitelink from '
                       f'{page.title(as_link=True, insite=self.repo)}')

        if data is None:
            data = {}
        data.setdefault('sitelinks', {}).update({
            page.site.dbName(): {
                'site': page.site.dbName(),
                'title': page.title()
            }
        })
        data.setdefault('labels', {}).update({
            page.site.lang: {
                'language': page.site.lang,
                'value': page.title(without_brackets=page.namespace() == 0)
            }
        })
        pywikibot.info(f'Creating item for {page}...')
        item = pywikibot.ItemPage(page.site.data_repository())
        kwargs.setdefault('show_diff', False)
        result = self.user_edit_entity(item, data, summary=summary, **kwargs)
        if result:
            return item
        return None

    def treat_page(self) -> None:
        """Treat a page."""
        page: pywikibot.page.BasePage | None = self.current_page
        if self.use_from_page is True:
            try:
                item = pywikibot.ItemPage.fromPage(page)
            except NoPageError:
                item = None
        elif isinstance(page, pywikibot.ItemPage):
            item = page
            page = None
        else:
            # FIXME: Hack because 'is_data_repository' doesn't work if
            #        site is the APISite. See T85483
            assert page is not None
            data_site = page.site.data_repository()
            if (data_site.family == page.site.family
                    and data_site.code == page.site.code):
                is_item = page.namespace() == data_site.item_namespace.id
            else:
                is_item = False
            if is_item:
                item = pywikibot.ItemPage(data_site, page.title())
                page = None
            else:
                try:
                    item = pywikibot.ItemPage.fromPage(page)
                except NoPageError:
                    item = None
                if self.use_from_page is False:
                    _error(f'{page} is not in the item namespace but must'
                           ' be an item.')
                    return

        assert not (page is None and item is None)

        if not item and self.create_missing_item:
            item = self.create_item_for_page(page, asynchronous=False)

        if not item and not self.treat_missing_item:
            pywikibot.info(f"{page} doesn't have a Wikidata item.")
            return

        self.treat_page_and_item(page, item)

    def treat_page_and_item(self, page: pywikibot.page.BasePage,
                            item: pywikibot.page.ItemPage) -> None:
        """Treat page together with its item (if it exists).

        Must be implemented in subclasses.
        """
        raise NotImplementedError(f'Method {type(self).__name__}.'
                                  'treat_page_and_item() not implemented.')


set_interface(config.userinterface)

critical = redirect_func(pwb_logging.critical,
                         target_module='pywikibot',
                         since='9.2.0')
debug = redirect_func(_debug,
                      target_module='pywikibot',
                      since='9.2.0')
error = redirect_func(_error,
                      target_module='pywikibot',
                      since='9.2.0')
exception = redirect_func(_exception,
                          target_module='pywikibot',
                          since='9.2.0')
log = redirect_func(_log,
                    target_module='pywikibot',
                    since='9.2.0')
output = redirect_func(pwb_logging.output,
                       target_module='pywikibot',
                       since='9.2.0')
stdout = redirect_func(_stdout,
                       target_module='pywikibot',
                       since='9.2.0')
warning = redirect_func(_warning,
                        target_module='pywikibot',
                        since='9.2.0')

# NOTE: (T286348)
# Do not use ModuleDeprecationWrapper with this module.
# pywikibot.bot.ui would be wrapped through the ModuleDeprecationWrapper
# and a cannot be changed later. Use another depecation method instead
# (until T286348 has been solved somehow different).



"""Interface of various types of MediaWiki pages."""
#
# (C) Pywikibot team, 2022-2023
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from typing import Union

from pywikibot.page._basepage import BasePage
from pywikibot.page._category import Category
from pywikibot.page._filepage import FileInfo, FilePage
from pywikibot.page._links import BaseLink, Link, SiteLink, html2unicode
from pywikibot.page._page import Page
from pywikibot.page._revision import Revision
from pywikibot.page._user import User
from pywikibot.page._wikibase import (
    Claim,
    ItemPage,
    LexemeForm,
    LexemePage,
    LexemeSense,
    MediaInfo,
    Property,
    PropertyPage,
    WikibaseEntity,
    WikibasePage,
)
from pywikibot.site import BaseSite as _BaseSite


__all__ = (
    'BaseLink',
    'Link',
    'SiteLink',
    'BasePage',
    'Page',
    'FilePage',
    'Category',
    'User',
    'WikibasePage',
    'ItemPage',
    'LexemePage',
    'LexemeForm',
    'LexemeSense',
    'PropertyPage',
    'Property',
    'Claim',
    'FileInfo',
    'WikibaseEntity',
    'MediaInfo',
    'Revision',
    'html2unicode',
)

PageSourceType = Union[
    BaseLink,
    BasePage,
    _BaseSite,
]



"""Family module for Wikispore.

.. versionadded:: 4.1
"""
#
# (C) Pywikibot team, 2020-2023
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


class Family(family.Family):

    """Family class for Wikispore.

    .. versionadded:: 4.1
    """

    name = 'wikispore'
    langs = {
        'en': 'wikispore.wmflabs.org',
        'test': 'wikispore-test.wmflabs.org',
    }



"""Superset Query interface.

.. versionadded:: 9.2
"""
#
# (C) Pywikibot team, 2024-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from http import HTTPStatus
from textwrap import fill
from typing import TYPE_CHECKING, Any

import pywikibot
from pywikibot.comms import http
from pywikibot.exceptions import NoUsernameError, ServerError


if TYPE_CHECKING:
    from pywikibot.site import BaseSite


class SupersetQuery:

    """Superset Query class.

    This class allows to run SQL queries against wikimedia superset
    service.
    """

    def __init__(self,
                 schema_name: str | None = None,
                 site: BaseSite | None = None,
                 database_id: int | None = None) -> None:
        """Create superset endpoint with initial defaults.

        Either site OR schema_name is required. Site and schema_name are
        mutually exclusive. Database id will be retrieved automatically
        if needed.

        :param site: The mediawiki site to be queried
        :param schema_name: superset database schema name. Example value
            "enwiki_p"
        :param database_id: superset database id.
        :raises TypeError: if site and schema_name are both defined'
        """
        if site and schema_name:
            msg = 'Only one of schema_name and site parameters can be defined'
            raise TypeError(msg)

        # Validate database_id
        if database_id and not isinstance(database_id, int):
            msg = f'database_id should be integer, but got "{database_id}"'
            raise TypeError(msg)

        self.site = site
        self.schema_name = schema_name
        self.database_id = database_id

        self.connected = False
        self.last_response = None
        self.superset_url = 'https://superset.wmcloud.org'

    def login(self) -> bool:
        """Login to superset.

        Function logins first to meta.wikimedia.org and then OAUTH login
        to superset.wmcloud.org. Working login expects that the user has
        manually permitted the username to login to the superset.

        :raises NoUsernameError: if not not logged in.
        :raises ServerError: For other errors
        :return: True if user has been logged to superset
        """
        # superset uses meta for OAUTH authentication
        loginsite = pywikibot.Site('meta')
        if not loginsite.logged_in():
            loginsite.login()
        if not loginsite.logged_in():
            msg = 'User is not logged in on meta.wikimedia.org'
            raise NoUsernameError(msg)

        # Superset oauth login
        url = f'{self.superset_url}/login/mediawiki?next='
        self.last_response = http.fetch(url)

        # Test if uset has been successfully logged in
        url = f'{self.superset_url}/api/v1/me/'
        self.last_response = http.fetch(url)

        # Handle error cases
        if self.last_response.status_code == HTTPStatus.OK:
            self.connected = True
        elif self.last_response.status_code == HTTPStatus.UNAUTHORIZED:
            self.connected = False
            raise NoUsernameError(fill(
                'User not logged in. You need to log in to '
                'meta.wikimedia.org and give OAUTH permission. '
                'Open https://superset.wmcloud.org/login/ '
                'with browser to login and give permission.'
            ))
        else:
            self.connected = False
            status_code = self.last_response.status_code
            raise ServerError(f'Unknown error: {status_code}')

        return self.connected

    def get_csrf_token(self) -> str:
        """Get superset CSRF token.

        Method retrieves a CSRF token from the Superset service. If the
        instance is not connected, it attempts to log in first.

        :raises ServerError: For any http errors
        :return: CSRF token string
        """
        if not self.connected:
            self.login()

        # Load CSRF token
        url = f'{self.superset_url}/api/v1/security/csrf_token/'
        self.last_response = http.fetch(url)

        if self.last_response.status_code == HTTPStatus.OK:
            return self.last_response.json()['result']

        status_code = self.last_response.status_code
        raise ServerError(f'CSRF token error:  {status_code}')

    def get_database_id_by_schema_name(self, schema_name: str) -> int:
        """Get superset database_id using superset schema name.

        :param schema_name: superset database schema name. Example value
            "enwiki_p"
        :raises KeyError: If the database ID could found.
        :raises ServerError: For any other http errors
        :return: database id
        """
        if not self.connected:
            self.login()

        for database_id in range(1, 20):
            url = self.superset_url
            url += f'/api/v1/database/{database_id}/schemas/?q=(force:!f)'
            self.last_response = http.fetch(url)

            if self.last_response.status_code == HTTPStatus.OK:
                schemas = self.last_response.json()['result']
                if schema_name in schemas:
                    return database_id

            elif self.last_response.status_code == HTTPStatus.NOT_FOUND:
                break
            else:
                status_code = self.last_response.status_code
                raise ServerError(f'Unknown error: {status_code}')

        url = self.superset_url
        raise KeyError(f'Schema "{schema_name}" not found in {url}.')

    def merge_query_arguments(self,
                              database_id: int | None = None,
                              schema_name: str | None = None,
                              site: BaseSite = None) -> tuple(int, str):
        """Determine and validate the database_id and schema_name.

        :param database_id: The superset database ID.
        :param schema_name: The superset schema name.
        :param site: The target site
        :raises TypeError: if site and schema_name are both defined'
        :raises TypeError: If determined database_id is not an integer.
        :raises TypeError: If neither site nor schema_name is
            determined.
        :return: A tuple containing database_id and schema_name.
        """
        if site and schema_name:
            msg = 'Only one of schema_name and site parameters can be defined'
            raise TypeError(msg)

        # Determine schema_name
        if not schema_name:
            if site:
                schema_name = f'{site.dbName()}_p'
            elif self.schema_name:
                schema_name = self.schema_name
            elif self.site:
                schema_name = f'{self.site.dbName()}_p'

        # Determine database_id
        if not database_id:
            if self.database_id:
                database_id = int(self.database_id)
            else:
                database_id = self.get_database_id_by_schema_name(schema_name)

        # Validate database_id
        if not isinstance(database_id, int):
            msg = f'database_id should be integer, but got "{database_id}"'
            raise TypeError(msg)

        # Ensure either site or schema_name is provided
        if not (self.site or schema_name):
            raise TypeError('Either site or schema_name must be provided')

        return database_id, schema_name

    def query(self, sql: str,
              database_id: int | None = None,
              schema_name: str | None = None,
              site: BaseSite = None) -> list[Any]:
        """Execute SQL queries on Superset.

        :param sql: The SQL query to execute.
        :param database_id: The database ID.
        :param schema_name: The schema name.
        :raises RuntimeError: If the query execution fails.
        :return: The data returned from the query execution.
        """
        if not self.connected:
            self.login()

        token = self.get_csrf_token()

        headers = {
            'X-CSRFToken': token,
            'Content-Type': 'application/json',
            'referer': 'https://superset.wmcloud.org/sqllab/'
        }

        database_id, schema_name = self.merge_query_arguments(database_id,
                                                              schema_name,
                                                              site)

        sql_query_payload = {
            'database_id': database_id,
            'schema': schema_name,
            'sql': sql,
            'json': True,
            'runAsync': False,
        }

        url = f'{self.superset_url}/api/v1/sqllab/execute/'
        try:
            self.last_response = http.fetch(uri=url,
                                            json=sql_query_payload,
                                            method='POST',
                                            headers=headers)
            self.last_response.raise_for_status()
            json = self.last_response.json()
            return json['data']

        except Exception as e:
            raise RuntimeError(f'Failed to execute query: {e}')



"""Various i18n functions.

Helper functions for both the internal localization system and for
TranslateWiki-based translations.

By default messages are assumed to reside in a package called
'scripts.i18n'. In pywikibot 3+, that package is not packaged with
pywikibot, and pywikibot 3+ does not have a hard dependency on any i18n
messages. However, there are three user input questions in pagegenerators
which will use i18n messages if they can be loaded.

The default message location may be changed by calling
:py:obj:`set_message_package` with a package name. The package must contain an
__init__.py, and a message bundle called 'pywikibot' containing messages.
See :py:obj:`twtranslate` for more information on the messages.
"""
#
# (C) Pywikibot team, 2004-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import json
import os
import pkgutil
import re
from collections import abc, defaultdict
from contextlib import suppress
from pathlib import Path
from textwrap import fill
from typing import Any

import pywikibot
from pywikibot import __url__, config
from pywikibot.backports import (
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Match,
    Sequence,
    cache,
    removesuffix,
)
from pywikibot.plural import plural_rule


PLURAL_PATTERN = r'{{PLURAL:(?:%\()?([^\)]*?)(?:\)d)?\|(.*?)}}'

# Package name for the translation messages. The messages data must loaded
# relative to that package name. In the top of this package should be
# directories named after for each script/message bundle, and each directory
# should contain JSON files called <lang>.json
_messages_package_name = 'scripts.i18n'
# Flag to indicate whether translation messages are available
_messages_available = None

_LANG_TO_GROUP_NAME = defaultdict(str, {
    'aa': 'aa',
    'ab': 'ab',
    'ace': 'ace',
    'ady': 'kbd',
    'af': 'af',
    'ak': 'ak',
    'als': 'als',
    'an': 'an',
    'arc': 'arc',
    'arn': 'an',
    'ary': 'arc',
    'arz': 'arc',
    'as': 'as',
    'ast': 'an',
    'atj': 'atj',
    'av': 'ab',
    'ay': 'an',
    'azb': 'azb',
    'ba': 'ab',
    'bar': 'bar',
    'bat-smg': 'bat-smg',
    'bcl': 'bcl',
    'be': 'be',
    'be-tarask': 'be',
    'bh': 'bh',
    'bho': 'bh',
    'bi': 'bi',
    'bjn': 'ace',
    'bm': 'atj',
    'bpy': 'as',
    'br': 'atj',
    'bs': 'bs',
    'bug': 'ace',
    'bxr': 'ab',
    'ca': 'ca',
    'cbk-zam': 'cbk-zam',
    'cdo': 'cdo',
    'ce': 'ab',
    'ceb': 'bcl',
    'ckb': 'ckb',
    'co': 'co',
    'crh': 'crh',
    'crh-latn': 'crh',
    'cs': 'cs',
    'csb': 'csb',
    'cu': 'cu',
    'cv': 'ab',
    'da': 'da',
    'diq': 'diq',
    'dsb': 'dsb',
    'dty': 'dty',
    'eml': 'eml',
    'eu': 'eu',
    'ext': 'an',
    'fab': 'fab',
    'ff': 'atj',
    'fit': 'fit',
    'fiu-vro': 'fiu-vro',
    'fo': 'fo',
    'frp': 'co',
    'frr': 'bar',
    'fur': 'eml',
    'fy': 'af',
    'gag': 'gag',
    'gan': 'cdo',
    'gl': 'gl',
    'glk': 'glk',
    'gn': 'gl',
    'grc': 'grc',
    'gsw': 'als',
    'hak': 'cdo',
    'hmo': 'meu',
    'hr': 'bs',
    'hsb': 'dsb',
    'ht': 'atj',
    'ia': 'ia',
    'id': 'ace',
    'ie': 'ia',
    'ii': 'cdo',
    'ik': 'ik',
    'ilo': 'bcl',
    'inh': 'ab',
    'io': 'io',
    'is': 'fo',
    'iu': 'ik',
    'jv': 'ace',
    'kaa': 'kaa',
    'kab': 'kab',
    'kbd': 'kbd',
    'kbp': 'atj',
    'kg': 'atj',
    'kj': 'kj',
    'kk': 'ab',
    'kl': 'kl',
    'koi': 'ab',
    'krc': 'ab',
    'ksh': 'bar',
    'ku': 'diq',
    'kv': 'ab',
    'ky': 'ab',
    'lad': 'an',
    'lb': 'lb',
    'lbe': 'ab',
    'lez': 'ab',
    'li': 'af',
    'lij': 'eml',
    'liv': 'liv',
    'lmo': 'eml',
    'ln': 'atj',
    'lrc': 'azb',
    'ltg': 'ltg',
    'lzh': 'zh-classical',
    'mai': 'mai',
    'map-bms': 'map-bms',
    'mdf': 'ab',
    'meu': 'meu',
    'mg': 'atj',
    'mhr': 'ab',
    'min': 'min',
    'minnan': 'zh-classical',
    'mk': 'cu',
    'mn': 'ab',
    'mo': 'mo',
    'mrj': 'ab',
    'ms': 'ace',
    'mwl': 'fab',
    'myv': 'ab',
    'mzn': 'glk',
    'nah': 'an',
    'nan': 'zh-classical',
    'nap': 'eml',
    'nb': 'no',
    'nds': 'nds',
    'nds-nl': 'nds-nl',
    'ne': 'ne',
    'new': 'ne',
    'ng': 'kj',
    'nn': 'nn',
    'no': 'no',
    'nov': 'io',
    'nrm': 'atj',
    'nso': 'nso',
    'nv': 'an',
    'oc': 'oc',
    'olo': 'olo',
    'os': 'ab',
    'pag': 'bcl',
    'pam': 'bcl',
    'pap': 'af',
    'pcd': 'atj',
    'pdc': 'bar',
    'pfl': 'bar',
    'pms': 'eml',
    'pnt': 'grc',
    'ps': 'azb',
    'pt': 'pt',
    'pt-br': 'pt',
    'qu': 'an',
    'rm': 'rm',
    'rmy': 'mo',
    'roa-rup': 'roa-rup',
    'roa-tara': 'eml',
    'rue': 'rue',
    'rup': 'roa-rup',
    'rw': 'atj',
    'sa': 'mai',
    'sah': 'ab',
    'sc': 'eml',
    'scn': 'eml',
    'se': 'se',
    'sg': 'atj',
    'sgs': 'bat-smg',
    'sh': 'bs',
    'sk': 'cs',
    'sli': 'sli',
    'so': 'arc',
    'sr': 'sr',
    'srn': 'af',
    'st': 'nso',
    'stq': 'stq',
    'su': 'ace',
    'sv': 'da',
    'szl': 'csb',
    'tcy': 'tcy',
    'tet': 'fab',
    'tg': 'ab',
    'ti': 'aa',
    'tpi': 'bi',
    'tt': 'tt',
    'tw': 'ak',
    'ty': 'atj',
    'tyv': 'ab',
    'udm': 'ab',
    'uk': 'ab',
    'vec': 'eml',
    'vep': 'vep',
    'vls': 'af',
    'vro': 'fiu-vro',
    'wa': 'atj',
    'war': 'bcl',
    'wo': 'atj',
    'wuu': 'cdo',
    'xal': 'ab',
    'xmf': 'xmf',
    'yi': 'yi',
    'yua': 'an',
    'yue': 'cdo',
    'za': 'cdo',
    'zea': 'af',
    'zh': 'zh-classical',
    'zh-classical': 'zh-classical',
    'zh-cn': 'cdo',
    'zh-hans': 'zh-classical',
    'zh-min-nan': 'zh-min-nan',
    'zh-tw': 'zh-classical',
    'zh-yue': 'cdo'})

_GROUP_NAME_TO_FALLBACKS: dict[str, list[str]] = {
    '': [],
    'aa': ['am'],
    'ab': ['ru'],
    'ace': ['id', 'ms', 'jv'],
    'af': ['nl'],
    'ak': ['ak', 'tw'],
    'als': ['als', 'gsw', 'de'],
    'an': ['es'],
    'arc': ['ar'],
    'as': ['bn'],
    'atj': ['fr'],
    'azb': ['fa'],
    'bar': ['de'],
    'bat-smg': ['bat-smg', 'sgs', 'lt'],
    'bcl': ['tl'],
    'be': ['be', 'be-tarask', 'ru'],
    'bh': ['bh', 'bho'],
    'bi': ['bi', 'tpi'],
    'bs': ['sh', 'hr', 'bs', 'sr', 'sr-el'],
    'ca': ['oc', 'es'],
    'cbk-zam': ['es', 'tl'],
    'cdo': ['zh', 'zh-hanszh-cn', 'zh-tw', 'zh-classical', 'lzh'],
    'ckb': ['ku'],
    'co': ['fr', 'it'],
    'crh': ['crh', 'crh-latn', 'uk', 'ru'],
    'cs': ['cs', 'sk'],
    'csb': ['pl'],
    'cu': ['bg', 'sr', 'sh'],
    'da': ['da', 'no', 'nb', 'sv', 'nn'],
    'diq': ['ku', 'ku-latn', 'tr'],
    'dsb': ['hsb', 'dsb', 'de'],
    'dty': ['ne'],
    'eml': ['it'],
    'eu': ['es', 'fr'],
    'fab': ['pt'],
    'fit': ['fi', 'sv'],
    'fiu-vro': ['fiu-vro', 'vro', 'et'],
    'fo': ['da', 'no', 'nb', 'nn', 'sv'],
    'gag': ['tr'],
    'gl': ['es', 'pt'],
    'glk': ['glk', 'mzn', 'fa', 'ar'],
    'grc': ['el'],
    'ia': ['ia', 'la', 'it', 'fr', 'es'],
    'ik': ['iu', 'kl'],
    'io': ['eo'],
    'kaa': ['uz', 'ru'],
    'kab': ['ar', 'fr'],
    'kbd': ['kbd', 'ady', 'ru'],
    'kj': ['kj', 'ng'],
    'kl': ['da', 'iu', 'no', 'nb'],
    'lb': ['de', 'fr'],
    'liv': ['et', 'lv'],
    'ltg': ['lv'],
    'mai': ['hi'],
    'map-bms': ['jv', 'id', 'ms'],
    'meu': ['meu', 'hmo'],
    'min': ['id'],
    'mo': ['ro'],
    'nds': ['nds-nl', 'de'],
    'nds-nl': ['nds', 'nl'],
    'ne': ['ne', 'new', 'hi'],
    'nn': ['no', 'nb', 'sv', 'da'],
    'no': ['no', 'nb', 'da', 'nn', 'sv'],
    'nso': ['st', 'nso'],
    'oc': ['fr', 'ca', 'es'],
    'olo': ['fi'],
    'pt': ['pt', 'pt-br'],
    'rm': ['de', 'it'],
    'roa-rup': ['roa-rup', 'rup', 'ro'],
    'rue': ['uk', 'ru'],
    'se': ['sv', 'no', 'nb', 'nn', 'fi'],
    'sli': ['de', 'pl'],
    'sr': ['sr-el', 'sh', 'hr', 'bs'],
    'stq': ['nds', 'de'],
    'tcy': ['kn'],
    'tt': ['tt-cyrl', 'ru'],
    'vep': ['et', 'fi', 'ru'],
    'xmf': ['ka'],
    'yi': ['he', 'de'],
    'zh-classical': ['zh', 'zh-hans', 'zh-tw', 'zh-cn', 'zh-classical', 'lzh'],
    'zh-min-nan': [
        'cdo', 'zh', 'zh-hans', 'zh-tw', 'zh-cn', 'zh-classical', 'lzh']
}


def set_messages_package(package_name: str) -> None:
    """Set the package name where i18n messages are located."""
    global _messages_package_name
    global _messages_available
    _messages_package_name = package_name
    _messages_available = None


def messages_available() -> bool:
    """Return False if there are no i18n messages available.

    To determine if messages are available, it looks for the package name
    set using :py:obj:`set_messages_package` for a message bundle called
    ``pywikibot`` containing messages.

    >>> from pywikibot import i18n
    >>> i18n.messages_available()
    True
    >>> old_package = i18n._messages_package_name  # save the old package name
    >>> i18n.set_messages_package('foo')
    >>> i18n.messages_available()
    False
    >>> i18n.set_messages_package(old_package)
    >>> i18n.messages_available()
    True
    """
    global _messages_available
    if _messages_available is not None:
        return _messages_available

    try:
        mod = __import__(_messages_package_name, fromlist=['__path__'])
    except ImportError:
        _messages_available = False
        return False

    _messages_available = bool(os.listdir(next(iter(mod.__path__))))
    return _messages_available


def _altlang(lang: str) -> list[str]:
    """Define fallback languages for particular languages.

    If no translation is available to a specified language, translate() will
    try each of the specified fallback languages, in order, until it finds
    one with a translation, with 'en' and '_default' as a last resort.

    For example, if for language 'xx', you want the preference of languages
    to be: xx > fr > ru > en, you let this method return ['fr', 'ru'].

    This code is used by other translating methods below.

    :param lang: The language code
    :return: language codes
    """
    return _GROUP_NAME_TO_FALLBACKS[_LANG_TO_GROUP_NAME[lang]]


@cache
def _get_bundle(lang: str, dirname: str) -> dict[str, str]:
    """Return json data of certain bundle if exists.

    For internal use, don't use it directly.

    .. versionadded:: 7.0
    """
    filename = f'{dirname}/{lang}.json'
    try:
        data = pkgutil.get_data(_messages_package_name, filename)
        assert data is not None
        trans_text = data.decode('utf-8')
    except OSError:  # file open can cause several exceptions
        return {}

    return json.loads(trans_text)


def _get_translation(lang: str, twtitle: str) -> str | None:
    """Return message of certain twtitle if exists.

    For internal use, don't use it directly.
    """
    message_bundle = twtitle.split('-')[0]
    transdict = _get_bundle(lang, message_bundle)
    return transdict.get(twtitle)


def _extract_plural(lang: str, message: str, parameters: Mapping[str, int]
                    ) -> str:
    """Check for the plural variants in message and replace them.

    :param message: the message to be replaced
    :param parameters: plural parameters passed from other methods
    :return: The message with the plural instances replaced
    """
    def static_plural_value(n: int) -> int:
        plural_rule = rule['plural']
        assert not callable(plural_rule)
        return plural_rule

    def replace_plural(match: Match[str]) -> str:
        selector = match[1]
        variants = match[2]
        num = parameters[selector]
        if not isinstance(num, int):
            raise ValueError(f"'{selector}' must be a number, not a {num} "
                             f'({type(num).__name__})')

        plural_entries = []
        specific_entries = {}
        # A plural entry cannot start at the end of the variants list,
        # and must end with | or the end of the variants list.
        for number, plural in re.findall(
            r'(?!$)(?: *(\d+) *= *)?(.*?)(?:\||$)', variants
        ):
            if number:
                specific_entries[int(number)] = plural
            else:
                assert not specific_entries, (
                    f'generic entries defined after specific in "{variants}"')
                plural_entries.append(plural)

        if num in specific_entries:
            return specific_entries[num]

        assert callable(plural_value)

        index = plural_value(num)
        needed = rule['nplurals']
        if needed == 1:
            assert index == 0

        if index >= len(plural_entries):
            # take the last entry in that case, see
            # https://translatewiki.net/wiki/Plural#Plural_syntax_in_MediaWiki
            index = -1
        return plural_entries[index]

    assert isinstance(parameters, Mapping), \
        f'parameters is not Mapping but {type(parameters)}'

    rule = plural_rule(lang)

    if callable(rule['plural']):
        plural_value = rule['plural']
    else:
        assert rule['nplurals'] == 1
        plural_value = static_plural_value

    return re.sub(PLURAL_PATTERN, replace_plural, message)


class _PluralMappingAlias(abc.Mapping):

    """Aliasing class to allow non mappings in _extract_plural.

    That function only uses __getitem__ so this is only implemented
    here.
    """

    def __init__(
        self,
        source: int | str | Sequence[int] | Mapping[str, int],
    ) -> None:
        self.source = source
        if isinstance(source, str):
            self.source = int(source)

        self.index = -1
        super().__init__()

    def __getitem__(self, key: str) -> int:
        self.index += 1
        if isinstance(self.source, dict):
            return int(self.source[key])

        if isinstance(self.source, (tuple, list)):
            if self.index < len(self.source):
                return int(self.source[self.index])
            raise ValueError('Length of parameter does not match PLURAL '
                             'occurrences.')
        assert isinstance(self.source, int)
        return self.source

    def __iter__(self) -> Iterator[int]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


DEFAULT_FALLBACK = ('_default', )


def translate(code: str | pywikibot.site.BaseSite,
              xdict: str | Mapping[str, Any],
              parameters: Mapping[str, int] | None = None,
              fallback: bool | Iterable[str] = False) -> Any | None:
    """Return the most appropriate localization from a localization dict.

    Given a site code and a dictionary, returns the dictionary's value
    for key *code* if this key exists; otherwise tries to return a value
    for an alternative code that is most applicable to use on the wiki
    in language *code* except fallback is False.

    The *code* itself is always checked first, then these codes that have
    been defined to be alternatives, and finally English.

    If *fallback* is False and the code is not found in the *xdict*,
    None is returned.

    For PLURAL support have a look at the :func:`twtranslate` function.

    .. versionchanged:: 2.0
       *parameter* other than a mapping (or None) is deprecated.
    .. versionchanged:: 6.2
       ValueError is raised if *parameter* is not a mapping.
    .. versionchanged:: 10.2
       TypeError instead of ValueError is raised if *parameter* is not a
       mapping.

    :param code: The site code as string or Site object. If xdict is an
        extended dictionary the Site object should be used in favour of
        the code string. Otherwise localizations from a wrong family
        might be used.
    :param xdict: dictionary with language codes as keys or extended
        dictionary with family names as keys containing code
        dictionaries or a single string. May contain PLURAL tags as
        described in twtranslate
    :param parameters: For passing (plural) parameters
    :param fallback: Try an alternate language code. If it's iterable
        it'll also try those entries and choose the first match.
    :return: the localized value, usually a string
    :raise IndexError: If the language supports and requires more
        plurals than defined for the given PLURAL pattern.
    :raise KeyError: No fallback key found if fallback is not False
    :raise TypeError: *parameter* is not a mapping
    """
    family = pywikibot.config.family
    # If a site is given instead of a code, use its code
    if hasattr(code, 'code'):
        family = code.family.name
        code = code.code
    assert isinstance(code, str)

    try:
        lookup = xdict[code]
    except (KeyError, TypeError):
        # Check whether xdict has multiple projects
        if isinstance(xdict, dict) and family in xdict:
            lookup = xdict[family]
        else:
            lookup = xdict

    # Get the translated string
    if not isinstance(lookup, dict):
        trans = lookup
    elif not lookup:
        trans = None
    else:
        codes = [code]
        if fallback is True:
            codes += [*_altlang(code), '_default', 'en']
        elif fallback is not False:
            assert not isinstance(fallback, bool)
            codes.extend(fallback)
        for code in codes:
            if code in lookup:
                trans = lookup[code]
                break
        else:
            if fallback is not False:
                raise KeyError(
                    f'No fallback key found in lookup dict for "{code}"')
            trans = None

    if trans is None:
        if isinstance(xdict, dict) and 'wikipedia' in xdict:
            # fallback to wikipedia family
            return translate(code, xdict['wikipedia'],
                             parameters=parameters, fallback=fallback)
        return None  # return None if we have no translation found

    if parameters is None:
        return trans

    if not isinstance(parameters, Mapping):
        raise TypeError(
            f'parameters should be a mapping, not {type(parameters).__name__}'
        )

    # else we check for PLURAL variants
    trans = _extract_plural(code, trans, parameters)
    if parameters:
        # On error: parameter is for PLURAL variants only,
        # don't change the string
        with suppress(KeyError, TypeError):
            trans = trans % parameters
    return trans


def get_bot_prefix(
    source: str | pywikibot.site.BaseSite,
    use_prefix: bool
) -> str:
    """Get the bot prefix string like 'Bot: ' including space delimiter.

    .. note:: If *source* is a str and ``config.bot_prefix`` is set to
       None, it cannot be determined whether the current user is a bot
       account. In this cas the prefix will be returned.
    .. versionadded:: 8.1

    :param source: When it's a site it's using the lang attribute and otherwise
        it is using the value directly.
    :param use_prefix: If True, return a bot prefix which depends on the
        ``config.bot_prefix`` setting.
    """
    config_prefix = config.bot_prefix_summary
    if not use_prefix or config_prefix is False:
        return ''

    if isinstance(config_prefix, str):
        return config_prefix + ' '

    try:
        prefix = twtranslate(source, 'pywikibot-bot-prefix') + ' '
    except pywikibot.exceptions.TranslationError:
        # the 'pywikibot' package is available but the message key may
        # be missing
        prefix = 'Bot: '

    if config_prefix is True \
       or not hasattr(source, 'lang') \
       or source.isBot(source.username()):
        return prefix

    return ''


def twtranslate(
    source: str | pywikibot.site.BaseSite,
    twtitle: str,
    parameters: Sequence[str] | Mapping[str, int] | None = None,
    *,
    fallback: bool = True,
    fallback_prompt: str | None = None,
    only_plural: bool = False,
    bot_prefix: bool = False
) -> str | None:
    r"""Translate a message using JSON files in messages_package_name.

    fallback parameter must be True for i18n and False for L10N or testing
    purposes.

    Support for plural is implemented like in MediaWiki extension. If the
    TranslateWiki message contains a plural tag inside which looks like::

        {{PLURAL:<number>|<variant1>|<variant2>[|<variantn>]}}

    it takes that variant calculated by the plural_rules depending on the
    number value. Multiple plurals are allowed.

    As an examples, if we had several json dictionaries in test folder like:

    en.json::

      {
         "test-plural": "Bot: Changing %(num)s {{PLURAL:%(num)d|page|pages}}.",
      }

    fr.json::

      {
         "test-plural": \
         "Robot: Changer %(descr)s {{PLURAL:num|une page|quelques pages}}.",
      }

    >>> import os
    >>> os.environ['PYWIKIBOT_TEST_QUIET'] = '1'
    >>> from pywikibot import i18n
    >>> i18n.set_messages_package('tests.i18n')
    >>> # use a dictionary
    >>> str(i18n.twtranslate('en', 'test-plural', {'num':2}))
    'Bot: Changing 2 pages.'
    >>> # use additional format strings
    >>> str(i18n.twtranslate(
    ...    'fr', 'test-plural', {'num': 1, 'descr': 'seulement'}))
    'Robot: Changer seulement une page.'
    >>> # use parameter for plural and format strings outside
    >>> str(i18n.twtranslate(
    ...    'fr', 'test-plural', {'num': 10}, only_plural=True
    ... ) % {'descr': 'seulement'})
    'Robot: Changer seulement quelques pages.'

    .. versionchanged:: 8.1
       the *bot_prefix* parameter was added.

    .. versionchanged:: 10.5
       *fallback_prompt* is now returned whenever no translation is found,
       including unknown keys in existing packages.

    :param source: When it's a site it's using the lang attribute and
        otherwise it is using the value directly. The site object is
        recommended.
    :param twtitle: The TranslateWiki string title, in <package>-<key>
        format
    :param parameters: For passing parameters. It should be a mapping
        but for backwards compatibility can also be a list, tuple or a
        single value. They are also used for plural entries in which
        case they must be a Mapping and will cause a TypeError otherwise.
    :param fallback: Try an alternate language code
    :param fallback_prompt: The English message if i18n is not available
    :param only_plural: Define whether the parameters should be only
        applied to plural instances. If this is False it will apply the
        parameters also to the resulting string. If this is True the
        placeholders must be manually applied afterwards.
    :param bot_prefix: If True, prepend the message with a bot prefix
        which depends on the ``config.bot_prefix`` setting
    :raise IndexError: If the language supports and requires more
        plurals than defined for the given translation template.
    :raise TypeError: If parameters are not a mapping for plural
        messages.
    :raise ValueError: If parameters are not a mapping but required.
    :raise TranslationError: If no translation found and
        *fallback_prompt* is None.
    """
    def _return_fallback_or_raise() -> str:
        """Return formatted fallback_prompt, or raise TranslationError."""
        if fallback_prompt is not None:
            if parameters and not only_plural:
                return prefix + fallback_prompt % parameters
            return prefix + fallback_prompt
        raise pywikibot.exceptions.TranslationError(
            fill(
                f'No translation available for key {twtitle} of '
                f'{_messages_package_name} package in language '
                f'{getattr(source, "lang", source)}. It can happen due to an '
                f'outdated or missing i18n submodule or files. '
                f'See {__url__}/i18n.'
            )
        )

    # Get the bot prefix, if requested
    prefix = get_bot_prefix(source, use_prefix=bot_prefix)

    # If the messages package isn't available at all, use fallback_prompt
    if not messages_available():
        return _return_fallback_or_raise()

    # Determine language code from source
    lang = getattr(source, 'lang', source)

    # Prepare list of languages to try; fallback adds alternatives and English
    langs = [lang]
    if fallback:
        langs += [*_altlang(lang), 'en']

    # Try each language until a translation is found
    for alt in langs:
        trans = _get_translation(alt, twtitle)
        if trans:
            break
    else:
        # No translation found: return fallback_prompt if available
        return _return_fallback_or_raise()

    # Handle plural forms if present
    if '{{PLURAL:' in trans:
        if not isinstance(parameters, Mapping):
            raise TypeError('parameters must be a mapping.')
        trans = _extract_plural(alt, trans, parameters)

    # Validate parameters type for string formatting
    if parameters is not None and not isinstance(parameters, Mapping):
        raise ValueError(
            f'parameters should be a mapping, not {type(parameters).__name__}'
        )

    # Apply string formatting if requested
    if not only_plural and parameters:
        trans = trans % parameters

    # Return the final translation with bot prefix
    return prefix + trans


def twhas_key(source: str | pywikibot.site.BaseSite, twtitle: str) -> bool:
    """Check if a message has a translation in the specified language code.

    The translations are retrieved from i18n.<package>, based on the
    callers import table.

    No code fallback is made.

    :param source: When it's a site it's using the lang attribute and
        otherwise it is using the value directly.
    :param twtitle: The TranslateWiki string title, in <package>-<key>
        format
    """
    # If a site is given instead of a code, use its language
    lang = getattr(source, 'lang', source)
    transdict = _get_translation(lang, twtitle)
    return transdict is not None


def twget_keys(twtitle: str) -> list[str]:
    """Return all language codes for a special message.

    :param twtitle: The TranslateWiki string title, in <package>-<key>
        format
    :raises OSError: the package i18n cannot be loaded
    """
    # obtain the directory containing all the json files for this package
    package = twtitle.split('-')[0]
    mod = __import__(_messages_package_name, fromlist=['__file__'])
    pathname = os.path.join(next(iter(mod.__path__)), package)

    # build a list of languages in that directory
    langs = [removesuffix(filename, '.json')
             for filename in sorted(os.listdir(pathname))
             if filename.endswith('.json')]

    # exclude languages does not have this specific message in that package
    # i.e. an incomplete set of translated messages.
    return [lang for lang in langs
            if lang != 'qqq' and _get_translation(lang, twtitle)]


def bundles(stem: bool = False) -> Generator[Path | str, None, None]:
    """A generator which yields message bundle names or its path objects.

    The bundle name usually corresponds with the script name which is
    localized.

    With ``stem=True`` the bundle names are given:

    >>> from pywikibot import i18n
    >>> bundles = sorted(i18n.bundles(stem=True))
    >>> len(bundles)
    40
    >>> bundles[:4]
    ['add_text', 'archivebot', 'basic', 'blockpageschecker']
    >>> bundles[-5:]
    ['unlink', 'unprotect', 'unusedfiles', 'weblinkchecker', 'welcome']
    >>> 'pywikibot' in bundles
    True

    With ``stem=False`` we get Path objects:

    >>> path = next(i18n.bundles())
    >>> path.is_dir()
    True
    >>> path.parent.as_posix()
    'scripts/i18n'

    .. versionadded:: 7.0

    :param stem: yield the Path.stem if True and the Path object otherwise
    """
    for dirpath in Path(*_messages_package_name.split('.')).iterdir():
        if dirpath.is_dir() and not dirpath.match('*__'):  # ignore cache
            if stem:
                yield dirpath.stem
            else:
                yield dirpath


def known_languages() -> list[str]:
    """Return all languages we have localizations for.

    >>> from pywikibot import i18n
    >>> i18n.known_languages()[:2]
    ['ab', 'aeb']
    >>> i18n.known_languages()[-10:]  # doctest: +SKIP
    ['vo', 'vro', 'wa', 'war', 'xal', 'xmf', 'yi', 'yo', 'yue', 'zh']
    >>> len(i18n.known_languages()) > 250
    True
    >>> 'ab' in i18n.known_languages()
    True
    >>> 'zh' in i18n.known_languages()
    True

    The implementation is roughly equivalent to:

    .. code-block:: Python

       langs = set()
       for dirpath in bundles():
           for fname in dirpath.iterdir():
               if fname.suffix == '.json':
                   langs.add(fname.stem)
        return sorted(langs)

    .. versionadded:: 7.0
    """
    return sorted(
        {fname.stem for dirpath in bundles() for fname in dirpath.iterdir()
         if fname.suffix == '.json'}
    )


def input(twtitle: str,
          parameters: Mapping[str, int] | None = None,
          password: bool = False,
          fallback_prompt: str | None = None) -> str:
    """Ask the user a question, return the user's answer.

    The prompt message is retrieved via :py:obj:`twtranslate` and uses the
    config variable 'userinterface_lang'.

    :param twtitle: The TranslateWiki string title, in <package>-<key> format
    :param parameters: The values which will be applied to the translated text
    :param password: Hides the user's input (for password entry)
    :param fallback_prompt: The English prompt if i18n is not available.
    """
    if messages_available():
        code = config.userinterface_lang
        prompt = twtranslate(code, twtitle, parameters)
    elif fallback_prompt:
        prompt = fallback_prompt
    else:
        raise pywikibot.exceptions.TranslationError(
            f'Unable to load messages package {_messages_package_name} for '
            f'bundle {twtitle}'
        )
    return pywikibot.input(prompt, password)


if not messages_available():
    set_messages_package('pywikibot.scripts.i18n')



"""Objects representing API interface to MediaWiki site extensions."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

import pywikibot
from pywikibot.backports import Generator, Iterable
from pywikibot.data import api
from pywikibot.echo import Notification
from pywikibot.exceptions import (
    APIError,
    Error,
    InconsistentTitleError,
    NoPageError,
    SiteDefinitionError,
)
from pywikibot.site._decorators import need_extension
from pywikibot.tools import merge_unique_dicts


if TYPE_CHECKING:
    from pywikibot.site import NamespacesDict


class BaseSiteProtocol(Protocol):
    _proofread_levels: dict[int, str]
    tokens: dict[str, str]

    def _generator(self, *args, **kwargs) -> api.Request:
        ...

    def _request(self, **kwargs) -> api.Request:
        ...

    def _update_page(self, *args, **kwargs) -> None:
        ...

    def encoding(self) -> str:
        ...

    @property
    def namespaces(self, **kwargs) -> NamespacesDict:
        ...

    def simple_request(self, **kwargs) -> api.Request:
        ...

    def querypage(
        self, *args, **kwargs
    ) -> Generator[tuple[pywikibot.Page, int], None, None]:
        ...


class EchoMixin:

    """APISite mixin for Echo extension."""

    @need_extension('Echo')
    def notifications(self, **kwargs):
        """Yield Notification objects from the Echo extension.

        :keyword str | None format: If specified, notifications will
            be returned formatted this way. Its value is either ``model``,
            ``special`` or ``None``. Default is ``special``.

        .. seealso:: :api:`Notifications` for other keywords.
        """
        params = {
            'action': 'query',
            'meta': 'notifications',
            'notformat': 'special',
        }

        for key, value in kwargs.items():
            params['not' + key] = value

        data = self.simple_request(**params).submit()
        notifications = data['query']['notifications']['list']

        return (Notification.fromJSON(self, notification)
                for notification in notifications)

    @need_extension('Echo')
    def notifications_mark_read(self: BaseSiteProtocol, **kwargs) -> bool:
        """Mark selected notifications as read.

        .. seealso:: :api:`echomarkread`

        :return: whether the action was successful
        """
        kwargs = merge_unique_dicts(kwargs, action='echomarkread',
                                    token=self.tokens['csrf'])
        req = self.simple_request(**kwargs)
        data = req.submit()
        try:
            return data['query']['echomarkread']['result'] == 'success'
        except KeyError:
            return False


class ProofreadPageMixin:

    """APISite mixin for ProofreadPage extension."""

    @need_extension('ProofreadPage')
    def _cache_proofreadinfo(self: BaseSiteProtocol, expiry=False) -> None:
        """Retrieve proofreadinfo from site and cache response.

        Applicable only to sites with ProofreadPage extension installed.

        The following info is returned by the query and cached:
        - self._proofread_index_ns: Index Namespace
        - self._proofread_page_ns: Page Namespace
        - self._proofread_levels: a dictionary with::

            keys: int in the range [0, 1, ..., 4]
            values: category name corresponding to the 'key' quality level
            e.g. on en.wikisource:

            .. code-block:: python

               {0: 'Without text', 1: 'Not proofread', 2: 'Problematic',
                3: 'Proofread', 4: 'Validated'}

        :param expiry: either a number of days or a datetime.timedelta object
        :type expiry: int (days), :py:obj:`datetime.timedelta`, False (config)
        :return: A tuple containing _proofread_index_ns,
            self._proofread_page_ns and self._proofread_levels.
        :rtype: Namespace, Namespace, dict
        """
        if (not hasattr(self, '_proofread_index_ns')
                or not hasattr(self, '_proofread_page_ns')
                or not hasattr(self, '_proofread_levels')):

            pirequest = self._request(
                expiry=pywikibot.config.API_config_expiry
                if expiry is False else expiry,
                parameters={'action': 'query', 'meta': 'proofreadinfo'}
            )

            pidata = pirequest.submit()
            ns_id = pidata['query']['proofreadnamespaces']['index']['id']
            self._proofread_index_ns = self.namespaces[ns_id]

            ns_id = pidata['query']['proofreadnamespaces']['page']['id']
            self._proofread_page_ns = self.namespaces[ns_id]

            self._proofread_levels = {}
            for ql in pidata['query']['proofreadqualitylevels']:
                self._proofread_levels[ql['id']] = ql['category']

    @property
    def proofread_index_ns(self):
        """Return Index namespace for the ProofreadPage extension."""
        if not hasattr(self, '_proofread_index_ns'):
            self._cache_proofreadinfo()
        return self._proofread_index_ns

    @property
    def proofread_page_ns(self):
        """Return Page namespace for the ProofreadPage extension."""
        if not hasattr(self, '_proofread_page_ns'):
            self._cache_proofreadinfo()
        return self._proofread_page_ns

    @property
    def proofread_levels(self):
        """Return Quality Levels for the ProofreadPage extension."""
        if not hasattr(self, '_proofread_levels'):
            self._cache_proofreadinfo()
        return self._proofread_levels

    @need_extension('ProofreadPage')
    def loadpageurls(self: BaseSiteProtocol,
                     page: pywikibot.page.BasePage) -> None:
        """Load URLs from api and store in page attributes.

        Load URLs to images for a given page in the "Page:" namespace.
        No effect for pages in other namespaces.

        .. versionadded:: 8.6

        .. seealso:: :api:`imageforpage`
        """
        title = page.title(with_section=False)
        # responsiveimages: server would try to render the other images as well
        # let's not load the server unless needed.
        prppifpprop = 'filename|size|fullsize'

        query = self._generator(api.PropertyGenerator,
                                type_arg='imageforpage',
                                titles=title.encode(self.encoding()),
                                prppifpprop=prppifpprop)
        self._update_page(page, query)


class GeoDataMixin:

    """APISite mixin for GeoData extension."""

    @need_extension('GeoData')
    def loadcoordinfo(self: BaseSiteProtocol, page) -> None:
        """Load [[mw:Extension:GeoData]] info."""
        title = page.title(with_section=False)
        query = self._generator(api.PropertyGenerator,
                                type_arg='coordinates',
                                titles=title.encode(self.encoding()),
                                coprop=['type', 'name', 'dim',
                                        'country', 'region',
                                        'globe'],
                                coprimary='all')
        self._update_page(page, query)


class PageImagesMixin:

    """APISite mixin for PageImages extension."""

    @need_extension('PageImages')
    def loadpageimage(self: BaseSiteProtocol, page) -> None:
        """Load [[mw:Extension:PageImages]] info.

        :param page: The page for which to obtain the image
        :type page: pywikibot.Page
        :raises APIError: PageImages extension is not installed
        """
        title = page.title(with_section=False)
        query = self._generator(api.PropertyGenerator,
                                type_arg='pageimages',
                                titles=title.encode(self.encoding()),
                                piprop=['name'])
        self._update_page(page, query)


class GlobalUsageMixin:

    """APISite mixin for Global Usage extension."""

    @need_extension('Global Usage')
    def globalusage(self, page, total=None):
        """Iterate global image usage for a given FilePage.

        :param page: the page to return global image usage for.
        :type page: pywikibot.FilePage
        :param total: iterate no more than this number of pages in
            total.
        :raises TypeError: input page is not a FilePage.
        :raises pywikibot.exceptions.SiteDefinitionError: Site could not
            be defined for a returned entry in API response.
        """
        if not isinstance(page, pywikibot.FilePage):
            raise TypeError(f'Page {page} must be a FilePage.')

        title = page.title(with_section=False)
        args = {'titles': title,
                'gufilterlocal': False,
                }
        query = self._generator(api.PropertyGenerator,
                                type_arg='globalusage',
                                guprop=['url', 'pageid', 'namespace'],
                                total=total,  # will set gulimit=total in api,
                                **args)

        for pageitem in query:
            if not self.sametitle(pageitem['title'],
                                  page.title(with_section=False)):
                raise InconsistentTitleError(page, pageitem['title'])

            api.update_page(page, pageitem, query.props)

            assert 'globalusage' in pageitem, \
                   "API globalusage response lacks 'globalusage' key"
            for entry in pageitem['globalusage']:
                try:
                    gu_site = pywikibot.Site(url=entry['url'])
                except SiteDefinitionError:
                    pywikibot.warning('Site could not be defined for global '
                                      f'usage for {page}: {entry}.')
                    continue
                gu_page = pywikibot.Page(gu_site, entry['title'])
                yield gu_page


class WikibaseClientMixin:

    """APISite mixin for WikibaseClient extension."""

    @need_extension('WikibaseClient')
    def unconnected_pages(
        self: BaseSiteProtocol,
        total: int | None = None,
        *,
        strict: bool = False
    ) -> Generator[pywikibot.Page, None, None]:
        """Yield Page objects from Special:UnconnectedPages.

        .. warning:: The retrieved pages may be connected in meantime.
           To avoid this, use *strict* parameter to check.

        .. versionchanged::
           The *strict* parameter was added.

        :param total: Maximum number of pages to return, or ``None`` for
            all.
        :param strict: If ``True``, verify that each page still has no
            data item before yielding it.
        """
        if total is not None and total <= 0:
            return

        if not strict:
            return self.querypage('UnconnectedPages', total)

        count = 0
        for page in self.querypage('UnconnectedPages'):
            if total is not None and count >= total:
                break

            try:
                page.data_item()
            except NoPageError:
                yield page
                count += 1


class LinterMixin:

    """APISite mixin for Linter extension."""

    @need_extension('Linter')
    def linter_pages(
        self: BaseSiteProtocol,
        lint_categories=None,
        total: int | None = None,
        namespaces=None,
        pageids: str | int | None = None,
        lint_from: str | int | None = None
    ) -> Iterable[pywikibot.Page]:
        """Return a generator to pages containing linter errors.

        :param lint_categories: categories of lint errors
        :type lint_categories: an iterable that returns values (str), or
            a pipe-separated string of values.
        :param total: if not None, yielding this many items in total
        :param namespaces: only iterate pages in these namespaces
        :type namespaces: iterable of str or Namespace key, or a single
            instance of those types. May be a '|' separated list of
            namespace identifiers.
        :param pageids: only include lint errors from the specified
            pageids
        :type pageids: an iterable that returns pageids, or a comma- or
             pipe-separated string of pageids (e.g. '945097,1483753,
             956608' or '945097|483753|956608')
        :param lint_from: Lint ID to start querying from
        :return: pages with Linter errors.
        """
        query = self._generator(api.ListGenerator, type_arg='linterrors',
                                total=total,  # Will set lntlimit
                                namespaces=namespaces)

        if lint_categories:
            if isinstance(lint_categories, str):
                lint_categories = lint_categories.split('|')
                lint_categories = [p.strip() for p in lint_categories]
            query.request['lntcategories'] = '|'.join(lint_categories)

        if pageids:
            if isinstance(pageids, str):
                pageids = pageids.split('|')
                pageids = [p.strip() for p in pageids]
            # Validate pageids.
            pageids = (str(int(p)) for p in pageids if int(p) > 0)
            query.request['lntpageid'] = '|'.join(pageids)

        if lint_from:
            query.request['lntfrom'] = int(lint_from)

        for pageitem in query:
            page = pywikibot.Page(self, pageitem['title'])
            api.update_page(page, pageitem)
            yield page


class ThanksMixin:

    """APISite mixin for Thanks extension."""

    @need_extension('Thanks')
    def thank_revision(self, revid, source=None):
        """Corresponding method to the 'action=thank' API action.

        :param revid: Revision ID for the revision to be thanked.
        :type revid: int
        :param source: A source for the thanking operation.
        :type source: str
        :raise APIError: On thanking oneself or other API errors.
        :return: The API response.
        """
        token = self.tokens['csrf']
        req = self.simple_request(action='thank', rev=revid, token=token,
                                  source=source)
        data = req.submit()
        if data['result']['success'] != 1:
            raise APIError('Thanking unsuccessful', '')
        return data


class UrlShortenerMixin:

    """APISite mixin for UrlShortener extension."""

    @need_extension('UrlShortener')
    def create_short_link(self, url):
        """Return a shortened link.

        Note that on Wikimedia wikis only metawiki supports this action,
        and this wiki can process links to all WM domains.

        :param url: The link to reduce, with propotol prefix.
        :type url: str
        :return: The reduced link, without protocol prefix.
        :rtype: str
        """
        req = self.simple_request(action='shortenurl', url=url)
        data = req.submit()
        return data['shortenurl']['shorturl']


class TextExtractsMixin:

    """APISite mixin for TextExtracts extension.

    .. versionadded:: 7.1
    """

    @need_extension('TextExtracts')
    def extract(self: BaseSiteProtocol,
                page: pywikibot.Page, *,
                chars: int | None = None,
                sentences: int | None = None,
                intro: bool = True,
                plaintext: bool = True) -> str:
        """Retrieve an extract of a page.

        :param page: The Page object for which the extract is read
        :param chars: How many characters to return.  Actual text
            returned might be slightly longer.
        :param sentences: How many sentences to return
        :param intro: Return only content before the first section
        :param plaintext: if True, return extracts as plain text instead
            of limited HTML

        .. seealso::

           - https://www.mediawiki.org/wiki/Extension:TextExtracts

           - :meth:`page.BasePage.extract`.
        """
        if not page.exists():
            raise NoPageError(page)
        req = self.simple_request(action='query',
                                  prop='extracts',
                                  titles=page.title(with_section=False),
                                  exchars=chars,
                                  exsentences=sentences,
                                  exintro=intro,
                                  explaintext=plaintext)
        data = req.submit()['query']['pages']
        if '-1' in data:
            msg = data['-1'].get('invalidreason',
                                 f"Unknown exception:\n{data['-1']}")
            raise Error(msg)

        return data[str(page.pageid)]['extract']



"""Miscellaneous helper functions for mysql queries."""
#
# (C) Pywikibot team, 2016-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot
from pywikibot import config


try:
    import pymysql
except ImportError:
    raise ImportError('MySQL python module not found. Please install PyMySQL.')


def mysql_query(query: str, params=None,
                dbname: str | None = None,
                verbose: bool | None = None):
    """Yield rows from a MySQL query.

    An example query that yields all ns0 pages might look like::

        SELECT
         page_namespace,
         page_title,
        FROM page
        WHERE page_namespace = 0;

    Supported MediaWiki projects use Unicode (UTF-8) character encoding.
    Cursor charset is utf8.

    :param query: MySQL query to execute
    :param params: input parameters for the query, if needed
        if list or tuple, %s shall be used as placeholder in the query string.
        if a dict, %(key)s shall be used as placeholder in the query string.
    :type params: tuple, list or dict of str
    :param dbname: db name
    :param verbose: if True, print query to be executed;
        if None, config.verbose_output will be used.
    :return: generator which yield tuples
    """
    # These are specified in config.py or your user config file
    if verbose is None:
        verbose = config.verbose_output > 0

    if config.db_connect_file is None:
        credentials = {'user': config.db_username,
                       'password': config.db_password}
    else:
        credentials = {'read_default_file': config.db_connect_file}

    args = {
        'host': config.db_hostname_format.format(dbname),
        'database': config.db_name_format.format(dbname),
        'port': config.db_port,
        'charset': 'utf8',
        'defer_connect': query == 'test',  # for tests
    }

    connection = pymysql.connect(**args, **credentials)

    with connection as conn, conn.cursor() as cursor:
        if verbose:
            _query = cursor.mogrify(query, params)

            if not isinstance(_query, str):
                _query = str(_query, encoding='utf-8')
            _query = _query.strip()
            _query = '\n'.join(f'    {line}'
                               for line in _query.splitlines())
            pywikibot.info('Executing query:\n' + _query)

        if query == 'test':  # for tests only
            yield query

        cursor.execute(query, params)
        yield from cursor



"""Object representing page revision."""
#
# (C) Pywikibot team, 2008-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import hashlib
from collections.abc import Mapping
from contextlib import suppress

from pywikibot import Timestamp


class Revision(Mapping):

    """A structure holding information about a single revision of a Page.

    Each data item can be accessed either by its key or as an attribute
    with the attribute name equal to the key e.g.:

    >>> r = Revision(comment='Sample for Revision access')
    >>> r.comment == r['comment']
    True
    >>> r.comment
    'Sample for Revision access'

    .. seealso::

       - :api:`Revisions`
       - :api:`Alldeletedrevisions`
    """

    def __init__(self, **kwargs) -> None:
        """Initializer."""
        self._data = kwargs
        self._upcast_dict(self._data)
        super().__init__()

    @staticmethod
    def _upcast_dict(map_) -> None:
        """Upcast dictionary values."""
        with suppress(KeyError):  # enable doctest
            map_['timestamp'] = Timestamp.fromISOformat(map_['timestamp'])

        map_.update(anon='anon' in map_)
        map_.update(minor='minor' in map_)
        map_.update(userhidden='userhidden' in map_)
        map_.update(commenthidden='commenthidden' in map_)

        map_.setdefault('comment', '')
        map_.setdefault('user', '')

        if 'slots' in map_:  # mw 1.32+
            mainslot = map_['slots'].get('main', {})
            map_['text'] = mainslot.get('*')
            map_['contentmodel'] = mainslot.get('contentmodel')
        else:
            map_['slots'] = None
            map_['text'] = map_.get('*')

        map_.setdefault('sha1')
        if map_['sha1'] is None and map_['text'] is not None:
            map_['sha1'] = hashlib.sha1(
                map_['text'].encode('utf8')).hexdigest()

    def __len__(self) -> int:
        """Return the number of data items."""
        return len(self._data)

    def __getitem__(self, name: str):
        """Return a single Revision item given by name."""
        if name in self._data:
            return self._data[name]

        return self.__missing__(name)

    # provide attribute access
    __getattr__ = __getitem__

    def __iter__(self):
        """Provide Revision data as iterator."""
        return iter(self._data)

    def __repr__(self) -> str:
        """String representation of Revision."""
        return f'{self.__class__.__name__}({self._data})'

    def __str__(self) -> str:
        """Printable representation of Revision data."""
        return str(self._data)

    def __missing__(self, key):
        """Provide backward compatibility for exceptions."""
        # raise AttributeError instead of KeyError for backward compatibility
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{key}'"
        )



"""Family module for Wikipedia."""
#
# (C) Pywikibot team, 2004-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


# The Wikimedia family that is known as Wikipedia, the Free Encyclopedia
class Family(family.SubdomainFamily, family.WikimediaFamily):

    """Family module for Wikipedia."""

    name = 'wikipedia'

    closed_wikis = [
        # https://noc.wikimedia.org/conf/highlight.php?file=dblists/closed.dblist
        'aa', 'ak', 'cho', 'ho', 'hz', 'ii', 'kj', 'kr', 'lrc', 'mh', 'mus',
        'na', 'ng', 'pih', 'ten',
    ]

    removed_wikis = [
        # https://noc.wikimedia.org/conf/highlight.php?file=dblists/deleted.dblist
        'dk', 'mo', 'ru-sib', 'tlh', 'tokipona', 'zh_cn', 'zh_tw',
    ]

    codes = {
        'ab', 'ace', 'ady', 'af', 'als', 'alt', 'am', 'ami', 'an', 'ang',
        'ann', 'anp', 'ar', 'arc', 'ary', 'arz', 'as', 'ast', 'atj', 'av',
        'avk', 'awa', 'ay', 'az', 'azb', 'ba', 'ban', 'bar', 'bat-smg', 'bbc',
        'bcl', 'bdr', 'be', 'be-tarask', 'bew', 'bg', 'bh', 'bi', 'bjn', 'blk',
        'bm', 'bn', 'bo', 'bpy', 'br', 'bs', 'btm', 'bug', 'bxr', 'ca',
        'cbk-zam', 'cdo', 'ce', 'ceb', 'ch', 'chr', 'chy', 'ckb', 'co', 'cr',
        'crh', 'cs', 'csb', 'cu', 'cv', 'cy', 'da', 'dag', 'de', 'dga', 'din',
        'diq', 'dsb', 'dtp', 'dty', 'dv', 'dz', 'ee', 'el', 'eml', 'en', 'eo',
        'es', 'et', 'eu', 'ext', 'fa', 'fat', 'ff', 'fi', 'fiu-vro', 'fj',
        'fo', 'fon', 'fr', 'frp', 'frr', 'fur', 'fy', 'ga', 'gag', 'gan',
        'gcr', 'gd', 'gl', 'glk', 'gn', 'gom', 'gor', 'got', 'gpe', 'gu',
        'guc', 'gur', 'guw', 'gv', 'ha', 'hak', 'haw', 'he', 'hi', 'hif', 'hr',
        'hsb', 'ht', 'hu', 'hy', 'hyw', 'ia', 'iba', 'id', 'ie', 'ig', 'igl',
        'ik', 'ilo', 'inh', 'io', 'is', 'it', 'iu', 'ja', 'jam', 'jbo', 'jv',
        'ka', 'kaa', 'kab', 'kbd', 'kbp', 'kcg', 'kg', 'kge', 'ki', 'kk', 'kl',
        'km', 'kn', 'knc', 'ko', 'koi', 'krc', 'ks', 'ksh', 'ku', 'kus', 'kv',
        'kw', 'ky', 'la', 'lad', 'lb', 'lbe', 'lez', 'lfn', 'lg', 'li', 'lij',
        'lld', 'lmo', 'ln', 'lo', 'lt', 'ltg', 'lv', 'mad', 'mai', 'map-bms',
        'mdf', 'mg', 'mhr', 'mi', 'min', 'mk', 'ml', 'mn', 'mni', 'mnw', 'mos',
        'mr', 'mrj', 'ms', 'mt', 'mwl', 'my', 'myv', 'mzn', 'nah', 'nap',
        'nds', 'nds-nl', 'ne', 'new', 'nia', 'nl', 'nn', 'no', 'nov', 'nqo',
        'nr', 'nrm', 'nso', 'nup', 'nv', 'ny', 'oc', 'olo', 'om', 'or', 'os',
        'pa', 'pag', 'pam', 'pap', 'pcd', 'pcm', 'pdc', 'pfl', 'pi', 'pl',
        'pms', 'pnb', 'pnt', 'ps', 'pt', 'pwn', 'qu', 'rki', 'rm', 'rmy', 'rn',
        'ro', 'roa-rup', 'roa-tara', 'rsk', 'ru', 'rue', 'rw', 'sa', 'sah',
        'sat', 'sc', 'scn', 'sco', 'sd', 'se', 'sg', 'sh', 'shi', 'shn', 'si',
        'simple', 'sk', 'skr', 'sl', 'sm', 'smn', 'sn', 'so', 'sq', 'sr',
        'srn', 'ss', 'st', 'stq', 'su', 'sv', 'sw', 'syl', 'szl', 'szy', 'ta',
        'tay', 'tcy', 'tdd', 'te', 'tet', 'tg', 'th', 'ti', 'tig', 'tk', 'tl',
        'tly', 'tn', 'to', 'tok', 'tpi', 'tr', 'trv', 'ts', 'tt', 'tum', 'tw',
        'ty', 'tyv', 'udm', 'ug', 'uk', 'ur', 'uz', 've', 'vec', 'vep', 'vi',
        'vls', 'vo', 'wa', 'war', 'wo', 'wuu', 'xal', 'xh', 'xmf', 'yi', 'yo',
        'za', 'zea', 'zgh', 'zh', 'zh-classical', 'zh-min-nan', 'zh-yue', 'zu',
    }

    # Sites we want to edit but not count as real languages
    test_codes = ['test', 'test2']

    # Templates that indicate a category redirect
    # Redirects to these templates are automatically included
    category_redirect_templates = {
        '_default': (),
        'ar': ('تحويل تصنيف',),
        'ary': ('Category redirect',),
        'arz': ('تحويل تصنيف',),
        'bn': ('বিষয়শ্রেণী পুনর্নির্দেশ',),
        'bs': ('Category redirect',),
        'ckb': ('ڕەوانەکەری پۆل', ),
        'cs': ('Zastaralá kategorie',),
        'da': ('Kategoriomdirigering',),
        'en': ('Category redirect',),
        'es': ('Categoría redirigida',),
        'eu': ('Kategoria birzuzendu',),
        'fa': ('رده بهتر',),
        'fr': ('Catégorie redirigée',),
        'gv': ('Aastiurey ronney',),
        'hi': ('श्रेणी अनुप्रेषित',),
        'hu': ('Kat-redir',),
        'id': ('Alih kategori',),
        'ja': ('Category redirect',),
        'ko': ('분류 넘겨주기',),
        'mk': ('Премести категорија',),
        'ml': ('Category redirect',),
        'ms': ('Pengalihan kategori',),
        'mt': ('Rindirizzament kategorija',),
        'ne': ('श्रेणी अनुप्रेषण',),
        'no': ('Kategoriomdirigering',),
        'pt': ('Redirecionamento de categoria',),
        'ro': ('Redirect categorie',),
        'ru': ('Переименованная категория',),
        'sco': ('Category redirect',),
        'sh': ('Prekat',),
        'simple': ('Category redirect',),
        'sl': ('Preusmeritev kategorije',),
        'sr': ('Category redirect',),
        'sq': ('Kategori e zhvendosur',),
        'sv': ('Kategoriomdirigering',),
        'tl': ('Category redirect',),
        'tr': ('Kategori yönlendirme',),
        'uk': ('Categoryredirect',),
        'ur': ('زمرہ رجوع مکرر',),
        'vi': ('Đổi hướng thể loại',),
        'yi': ('קאטעגאריע אריבערפירן',),
        'zh': ('分类重定向',),
        'zh-yue': ('分類彈去',),
    }

    # Global bot allowed languages on
    # https://meta.wikimedia.org/wiki/BPI#Current_implementation
    # & https://meta.wikimedia.org/wiki/Special:WikiSets/2
    cross_allowed = [
        'ab', 'ace', 'ady', 'af', 'als', 'am', 'ami', 'an', 'ang', 'ar', 'arc',
        'arz', 'as', 'ast', 'atj', 'av', 'avk', 'awa', 'ay', 'az', 'azb', 'ba',
        'ban', 'bar', 'bat-smg', 'bcl', 'be', 'be-tarask', 'bh', 'bi', 'bjn',
        'blk', 'bm', 'bo', 'bpy', 'bug', 'bxr', 'ca', 'cbk-zam', 'cdo', 'ce',
        'ceb', 'ch', 'chr', 'chy', 'ckb', 'co', 'cr', 'crh', 'cs', 'csb', 'cu',
        'cv', 'cy', 'dag', 'diq', 'dsb', 'dty', 'dz', 'ee', 'eml', 'eo', 'et',
        'eu', 'ext', 'fa', 'ff', 'fj', 'fo', 'frp', 'frr', 'fur', 'ga', 'gag',
        'gan', 'gcr', 'gd', 'glk', 'gn', 'gom', 'gor', 'got', 'gu', 'guc',
        'gur', 'guw', 'gv', 'ha', 'hak', 'haw', 'hi', 'hif', 'hr', 'hsb', 'ht',
        'hu', 'hy', 'hyw', 'ia', 'ie', 'ig', 'ik', 'ilo', 'inh', 'io', 'iu',
        'jam', 'jbo', 'jv', 'ka', 'kaa', 'kab', 'kbd', 'kbp', 'kcg', 'kg',
        'ki', 'kk', 'kl', 'km', 'kn', 'ko', 'koi', 'krc', 'ks', 'ku', 'kv',
        'kw', 'ky', 'la', 'lad', 'lbe', 'lez', 'lfn', 'lg', 'li', 'lij', 'lld',
        'lmo', 'ln', 'lo', 'lt', 'ltg', 'lv', 'mad', 'mai', 'map-bms', 'mdf',
        'mg', 'mi', 'min', 'mk', 'mn', 'mni', 'mrj', 'mwl', 'my', 'myv', 'mzn',
        'na', 'nah', 'nap', 'nds-nl', 'ne', 'new', 'nia', 'nl', 'nov', 'nqo',
        'nrm', 'nso', 'nv', 'ny', 'oc', 'olo', 'om', 'or', 'os', 'pa', 'pag',
        'pam', 'pap', 'pcd', 'pcm', 'pdc', 'pfl', 'pi', 'pih', 'pms', 'pnb',
        'pnt', 'ps', 'pwn', 'qu', 'rm', 'rmy', 'rn', 'roa-rup', 'roa-tara',
        'rue', 'rw', 'sa', 'sah', 'sat', 'sc', 'scn', 'sco', 'sd', 'se', 'sg',
        'sh', 'shi', 'shn', 'si', 'simple', 'sk', 'skr', 'sm', 'smn', 'sn',
        'so', 'srn', 'st', 'stq', 'su', 'sw', 'szl', 'szy', 'ta', 'tay', 'tcy',
        'te', 'tet', 'tg', 'th', 'ti', 'tk', 'tl', 'tn', 'to', 'tpi', 'trv',
        'ts', 'tt', 'tum', 'tw', 'ty', 'tyv', 'udm', 'ug', 'uz', 've', 'vec',
        'vep', 'vls', 'vo', 'wa', 'war', 'wo', 'xal', 'xh', 'xmf', 'yi', 'yo',
        'za', 'zea', 'zh', 'zh-classical', 'zh-min-nan', 'zh-yue', 'zu',
    ]

    # Languages that used to be coded in iso-8859-1
    latin1old = {
        'af', 'bs', 'co', 'cs', 'da', 'de', 'en', 'es', 'et', 'eu', 'fi', 'fr',
        'fy', 'ga', 'gl', 'ia', 'id', 'it', 'la', 'lt', 'lv', 'mi', 'mr', 'na',
        'nds', 'nl', 'no', 'pt', 'simple', 'sl', 'sv', 'sw', 'test', 'tt',
        'uk', 'vi', 'vo'
    }

    # Subpages for documentation.
    # TODO: List is incomplete, to be completed for missing languages.
    # TODO: Remove comments for appropriate pages
    doc_subpages = {
        '_default': (('/doc', ),
                     ['arz', 'bn', 'cs', 'da', 'en', 'es', 'hr', 'hu', 'id',
                      'ilo', 'ja', 'ms', 'pt', 'ro', 'ru', 'simple', 'sh',
                      'vi', 'zh']
                     ),
        'ar': ('/شرح', '/doc'),
        'ary': ('/توثيق', '/شرح', '/doc'),
        'bs': ('/dok', ),
        'ca': ('/ús', ),
        'de': ('Doku', '/Meta'),
        'dsb': ('/Dokumentacija', ),
        'eu': ('txantiloi dokumentazioa', '/dok'),
        'fa': ('/doc', '/توضیحات'),
        # fi: no idea how to handle this type of subpage at :Metasivu:
        'fi': ((), ),
        'fr': ('/Documentation',),
        'hsb': ('/Dokumentacija', ),
        'it': ('/Man', ),
        'ka': ('/ინფო', ),
        'ko': ('/설명문서', ),
        'no': ('/dok', ),
        'nn': ('/dok', ),
        'pl': ('/opis', ),
        'sk': ('/Dokumentácia', ),
        'sr': ('/док', ),
        'sv': ('/dok', ),
        'uk': ('/Документація', ),
        'ur': ('/doc', '/دستاویز'),
    }

    # Templates that indicate an edit should be avoided
    edit_restricted_templates = {
        'ar': ('تحرر',),
        'ary': ('كاتبدل دابا',),
        'arz': ('بتتطور',),
        'bs': ('Izmjena u toku',),
        'cs': ('Pracuje se',),
        'de': ('Inuse', 'In use', 'In bearbeitung', 'Inbearbeitung'),
        'en': ('Inuse', 'In use'),
        'fa': ('ویرایش',),
        'fr': ('En cours',),
        'he': ('בעבודה',),
        'hr': ('Radovi',),
        'hy': ('Խմբագրում եմ',),
        'ro': ('Dezvoltare', 'S-dezvoltare', 'Modific acum'),
        'ru': ('Редактирую',),
        'sr': ('Радови у току', 'Рут'),
        'test': ('In use',),
        'ur': ('زیر ترمیم',),
        'zh': ('Inuse',),
    }

    # Archive templates that indicate an edit of non-archive bots
    # should be avoided
    archived_page_templates = {
        'ar': ('أرشيف نقاش',),
        'arz': ('صفحة ارشيف',),
        'cs': ('Archiv', 'Archiv Wikipedie', 'Archiv diskuse',
               'Archivace start', 'Posloupnost archivů', 'Rfa-archiv-start',
               'Rfc-archiv-start'),
        'de': ('Archiv',),
        'ro': ('Arhivă',),
    }

    citoid_endpoint = '/api/rest_v1/data/citation/'

    @classmethod
    def __post_init__(cls) -> None:
        """Add 'yue' code alias due to :phab:`T341960`.

        .. versionadded:: 8.3
        """
        aliases = cls.code_aliases.copy()
        aliases['yue'] = 'zh-yue'
        cls.code_aliases = aliases

    def encodings(self, code):
        """Return a list of historical encodings for a specific site."""
        # Historic compatibility
        if code == 'pl':
            return 'utf-8', 'iso8859-2'
        if code == 'ru':
            return 'utf-8', 'iso8859-5'
        if code in self.latin1old:
            return 'utf-8', 'iso-8859-1'
        return super().encodings(code)
